vec3 plancks(in float t, in vec3 lambda) {
    const float h = 6.63e-16;
    const float c = 3.0e17;
    const float k = 1.38e-5;
    vec3 p1 = (2.0*h*pow(c, 2.0)) / pow(lambda, vec3(5.0));
    vec3 p2 = exp(h*c/(lambda*k*t))-vec3(1);
    return (p1 / p2) * pow(1e9, 2.0);
}

vec3 blackbody(in float t) {
    vec3 xyz = vec3(0.0);

    for (float w = 390.0; w < 830.0; w += 5.0)
    {
        vec3 wl = vec3(w);
        float r = plancks(t, wl).x;
        float g = plancks(t, wl).y;
        float b = plancks(t, wl).z;
        xyz = r * spectrum_to_xyz(wl.x) + xyz;
        xyz = g * spectrum_to_xyz(wl.y) + xyz;
        xyz = b * spectrum_to_xyz(wl.z) + xyz;
    }

    vec3 rgb = xyz_to_rgb(xyz);
        #ifndef TemperatureBrightness
         rgb = rgb / max(rgb.x, max(rgb.y, rgb.z));
        #endif

    return rgb;
}