vec2 bokeh_ghost_coord(vec2 light_pos, float flare_offset, float dist_from_focal, out float size) {
	//This function seems to be the best way to do this.
	float focus = abs(screenspace_to_viewspace(texture(colortex7, vec2(0.5)).a, gbufferProjectionInverse));
	#ifdef NoSizeChangeLensFlare
		focus = abs(screenspace_to_viewspace(1.0, gbufferProjectionInverse));
	#endif
    size = calculate_camera_coc(focus, abs(dist_from_focal + (0.5 * sensor_width * gbufferProjection[0].x))) / aspectRatio;
	size *= 0.25;
	vec2 aspect_correct = vec2(1.0, aspectRatio);

	vec2 coord = ((texture_coordinate * 2.0 - 1.0) / aspect_correct / size) * 0.5 + 0.5;
	vec2 flare_pos = (coord / pow(2.0, LensLOD) + bokeh_offset);

	flare_pos = flare_pos + ((light_pos * 2.0 - 1.0) / aspect_correct) * ((-flare_offset*0.25) / size);
	return flare_pos;
}

vec3 bokeh_ghost(vec2 light_pos, float flare_offset, float dist_from_focal) {
	float size;
	vec2 coord = bokeh_ghost_coord(light_pos, flare_offset, dist_from_focal, size);
	vec4 ghost = texture(colortex6, coord);
	return ghost.rgb * ghost.a;
}

float smooth_circle(in vec2 coord, float dist) {
	//This function is made by Sgt Sarcasm.
	coord  = coord * 2.0 - 1.0;
	coord *= dist;
	coord  = coord * 0.5 + 0.5;

	float xvect = coord.x * aspectRatio - texture_coordinate.x * aspectRatio;
	float yvect = coord.y - texture_coordinate.y;

	return sqrt(xvect*xvect + yvect*yvect);
}

float smooth_ghost(vec2 light_pos, float flare_offset, float dist_from_focal) {
	float size;
	float focus = abs(screenspace_to_viewspace(texture(colortex7, vec2(0.5)).r, gbufferProjectionInverse));
	#ifdef NoSizeChangeLensFlare
		focus = abs(screenspace_to_viewspace(1.0, gbufferProjectionInverse));
	#endif
    size = calculate_camera_coc(focus, abs(dist_from_focal + (0.5 * sensor_width * gbufferProjection[0].x))) / aspectRatio;
	size *= 0.25;
	return square(inversesqrt(smooth_circle(light_pos, flare_offset) * 10.0 + size + 0.001) * 0.5);
}

vec3 lens_flares() {
	//Currently not modeled after any particular camera lens system, but that will eventually change.
	//The values used here are also likely incorrect, but whatever, I think they look fine for now.
	vec4 tpos = vec4(sunPosition,1.0)*gbufferProjection;
			tpos = vec4(tpos.xyz/tpos.w,1.0);
			vec2 light_pos = tpos.xy/tpos.z;
			light_pos = (light_pos + 1.0f)/2.0f;

	float sun_vis = float(texture(depthtex1, light_pos).r >= 1.0);

	float distof = min(min(1.0-light_pos.x,light_pos.x),min(1.0-light_pos.y,light_pos.y));
	float fading = clamp(1.0-step(distof,0.1)+pow(distof*10.0,5.0),0.0,1.0);

	vec3 sun_color = texture(colortex0, light_pos).rgb / (MaxColorRange*32.0);

	vec3 lens_col = thin_film_interferance(21.00, ls.bk7_ior / ls.air_ior, radians(140.0));

	float dist_array[9] = float[](
		-0.25, //med-large
		-0.30, //large
		-0.35, //medium
		-0.40, //small
		-0.70, //small
		-0.80, //large
		-0.90, //small
		-0.95, //med
		-1.00  //small
	);

	float dist_from_focal_array[9] = float[](
		-3.0, //med-large
		 2.5, //large
		-4.0, //medium
		-5.0, //small
		-4.0, //small
		 2.5, //large
		 4.0, //small
		-3.0, //med
		-4.0  //small
	);

	float dist_array_smooth[2] = float[](
		-0.25,
		-0.90
	);

	float dist_from_focal_array_smooth[2] = float[](
		-3.0,
		-4.0
	);

	vec3 flare;

	//Flare loop
	for(int i = 0; i < 9; ++i) {
		flare += bokeh_ghost(light_pos, dist_array[i], dist_from_focal_array[i]);
	}

	//Smooth ghost loop
	for(int i = 0; i < 2; ++i) {
		flare += smooth_ghost(light_pos, dist_array_smooth[i], dist_from_focal_array_smooth[i]) / pi;
	}

	flare += smooth_ghost(light_pos, -0.95 / smooth_circle(light_pos, 0.0), 3.0) / pi;

	flare *= sun_color * fading * lens_col;

	return flare * sun_vis;
}