vec4 CalculateSphericalHarmonics(vec3 xyz) {
    const vec2 freqW = vec2(sqrt(1.0 / (4.0 * pi)), sqrt(3.0 / (4.0 * pi)));
    return vec4(freqW.x, freqW.y * xyz.yzx);
}

vec3[4] CalculateSphericalHarmonicCoefficients(vec3 value, vec3 xyz) {
    vec4 harmonics = CalculateSphericalHarmonics(xyz);
    return vec3[4](value * harmonics.x, value * harmonics.y, value * harmonics.z, value * harmonics.w);
}
vec3 ValueFromSphericalHarmonicCoefficients(vec3[4] coefficients, vec3 xyz) {
    vec4 harmonics = CalculateSphericalHarmonics(xyz);
    return coefficients[0] * harmonics.x + coefficients[1] * harmonics.y + coefficients[2] * harmonics.z + coefficients[3] * harmonics.w;
}

vec4[3] CalculateSphericalHarmonicCoefficientsAlt(vec3 value, vec3 xyz) {
    vec4 harmonics = CalculateSphericalHarmonics(xyz);
    return vec4[3](value.r * harmonics, value.g * harmonics, value.b * harmonics);
}
vec3 ValueFromSphericalHarmonicCoefficientsAlt(vec4[3] coefficients, vec3 xyz) {
    vec4 harmonics = CalculateSphericalHarmonics(xyz);
    return vec3(dot(coefficients[0], harmonics), dot(coefficients[1], harmonics), dot(coefficients[2], harmonics));
}

#include "/lib/distortion.glsl"

const vec2[12] offset = vec2[12](
	                 vec2(-0.5, 1.5), vec2( 0.5, 1.5),
	vec2(-1.5, 0.5), vec2(-0.5, 0.5), vec2( 0.5, 0.5), vec2( 1.5, 0.5),
	vec2(-1.5,-0.5), vec2(-0.5,-0.5), vec2( 0.5,-0.5), vec2( 1.5,-0.5),
	                 vec2(-0.5,-1.5), vec2( 0.5,-1.5)
);

const float bias_amount = 2e-5;

float shadow_blur_size = 1.5e-4;

uniform mat4 shadowProjectionInverse;

#include "shadow/shadows.glsl"

vec3 ggx_diffuse(in vec3 albedo, in float n_dot_v, in float n_dot_l, in float n_dot_h, in float l_dot_v, in float roughness) {
    float facing = 0.5 + 0.5 * l_dot_v;
    float rough = n_dot_h <= 0.0 ? 0.0 : facing * (0.9 - 0.4 * facing) * ((0.5 + n_dot_h) / n_dot_h);
    //This is named smooth_v to avoid any potential errors.
    float smooth_v = 1.05 * (1.0 - pow5((1.0 - n_dot_l))) * (1.0 - pow5((1.0 - n_dot_v)));
    float single = 1.0 * mix(smooth_v, rough, roughness) / pi;
    float multi = 0.1159 * roughness;

    return max((single + albedo * multi) * n_dot_l, 0.0);
}

#include "diffuse/oren_nayar.glsl"

/*
bool water_shadow(vec3 position, inout float water_depth, inout float fraction) {
    float shadowOffset = 8e-4;

    position.z /= SHADOW_DEPTH_SCALE;

	bool result = false;

	vec2 sample_coord = position.xy;
	     sample_coord = distort_shadowspace(sample_coord);
	     sample_coord = sample_coord * 0.5 + 0.5;

    position.z = position.z * 0.5 + 0.5;

    water_depth = SHADOW_DEPTH_RADIUS * max(position.z - texture(shadowtex0, sample_coord.xy).r, 0); //This is how it is done in Spectrum

    float water_cast = texture(shadowcolor1, sample_coord).r;

    float vis0 = step(position.z, texture(shadowtex0, sample_coord.xy).r);
    float vis1 = step(position.z, water_cast);
    fraction = vis1 - vis0 * vis1;

    if(water_cast > 0.9) result = true;

	return result;
}
*/

float calculate_blocklight_falloff(float block_light) {
    const float lightmap_range = 16.0;
    const float light_size     = 0.25;
    const float lightmap_scale = lightmap_range / light_size;

    const float cmp   = lightmap_scale * 0.5 - 0.5;
    const float sqMul = 4.0 / ((lightmap_scale + 1.0) * (lightmap_scale + 1.0));
    const float sqAdd = -sqMul * lightmap_scale;

    float falloff = lightmap_scale - lightmap_scale * block_light;
          falloff = falloff < cmp ? 1.0 / (falloff + 1.0) : sqMul * falloff + sqAdd;
    return falloff * falloff;
}

vec2 calculate_dither_pattern() {
    vec2 coord = texture_coordinate;
	coord *= view_size;
	coord = mod(coord + vec2(0.0), vec2(10.0));
	coord /= noiseTextureResolution;
    return texture(noisetex, coord).xy;
}

#if World == 0
    //#include "indirect/rsm.glsl"

    vec3 apply_shading(in vec3 col, in bool do_particles) {
        vec3 shading = vec3(0.0);

        float intensity = 2600.0;
        vec3 torch = blackbody(BlocklightTemperature) * intensity;

        vec3 skylight = vec3(0.0);
        #if IBL_Method == 0
            skylight = max(ValueFromSphericalHarmonicCoefficientsAlt(shcoeffs, sd.wnormal), 0.0) * mix(2.5, 2.0, abs(dot(sd.wnormal, up_vector)));
        #elif IBL_Method == 1
            for(int i = 0; i < 8; ++i) {
                vec3 dir = mat3(gbufferModelViewInverse) * fNormalize(gen_unit_vector(hash2(pd.view_position + i)));
                if(dot(sd.wnormal, dir) < 0.0) dir = -dir;
                skylight += max(texture(colortex3, project_sky(dir)).rgb * dot(sd.wnormal, dir), 0.0);
            }
            skylight /= 8;
            skylight = skylight;
        #endif
        #ifdef AtmosAnalytical
            skylight *= 0.65;
        #endif

        if(do_particles && texture(colortex0, texture_coordinate).a < 0.75) skylight = light.sky / 2.0;

        float id = texture(colortex2, texture_coordinate).a * 65535.0;

        vec3 h = normalize(normalize(-pd.view_position) + (mat3(gbufferModelView) * shadow_light_vector));
        float ndotv = dot(sd.vnormal, normalize(-pd.view_position));
        float ndotl = max(0.0, dot(sd.vnormal, mat3(gbufferModelView) * shadow_light_vector));
        float ndoth = dot(sd.vnormal, h);
        float ldotv = dot(mat3(gbufferModelView) * shadow_light_vector, normalize(-pd.view_position));
        float ldoth = dot(mat3(gbufferModelView) * shadow_light_vector, h);
        float ndotl2 = max(0.0, dot(mat3(gbufferModelView) * decodeNormal(decode2x8(texture(colortex5, texture_coordinate).r)), mat3(gbufferModelView) * shadow_light_vector));
        #ifdef MultNdotL2
            ndotl2 = ndotl;
        #endif

        bool subsurface = false;
        bool grass = false;
        bool leaves = false;
        if(id == 6.0 || id == 31.0 || id == 32.0 || id == 37.0 || id == 38.0 || id == 59.0 || id == 175.0) grass = true;
        if(id == 18.0 || id == 106.0 || id == 111.0 || id == 161.0) leaves = true;
        #ifdef FancySubsurface
            if(grass || leaves && ndotl <= 0.0) subsurface = true;
        #endif

        float cloud_shadow = cloud_shadow_2d(shadow_light_vector, pd.scene_position + cameraPosition);
        vec3 shadow_map = mix(calculate_shadow_map(pd.shadow_clip_position, subsurface), vec3(0.0), rainStrength);
        if(subsurface) shadow_map = vec3(0.0);

        float sss_phase = cornette_shanks_mie_phase(dot(fNormalize(pd.view_position), mat3(gbufferModelView) * shadow_light_vector), 0.75);
        sss_phase = (0.5 / pi) + sss_phase * 0.3;

        vec3 sss;
        if(!subsurface) {
            sss = vec3(0.0);
        } else {
            sss = light.direct * sss_phase;
            sss *= calculate_shadow_map(pd.shadow_clip_position, subsurface);
            sss *= float(subsurface);
            sss *= rcp(pi);
        }

        vec3 diffuse = ggx_diffuse(col * (1.0 - sd.f0), ndotv, ndotl, ndoth, ldotv, sd.roughness_squared) * ndotl2;
        if(subsurface) diffuse = ndotl * vec3(0.5);
        if(do_particles && texture(colortex0, texture_coordinate).a < 0.75) diffuse = vec3(0.2);
        #ifndef FancySubsurface
            if(ndotl <= 0.0 && texture(colortex0, texture_coordinate).a > 0.75) shadow_map = vec3(0.0);
        #endif

        vec3 shadow_view_normal = scenespace_to_shadowview(sd.wnormal, shadowModelView);

        //vec3 rsm = reflective_shadow_mapping(pd.shadow_view_position, pd.shadow_clip_position, shadow_view_normal, calculate_dither_pattern()) * light.direct;

        shading = (shadow_map * light.direct) * diffuse * (float(!subsurface)) * cloud_shadow + shading;
        //shading = rsm * cloud_shadow + shading;
        shading = calculate_blocklight_falloff(decode2x8(texture(colortex2, texture_coordinate).b).x) * torch + shading;
        shading = pow(decode2x8(texture(colortex2, texture_coordinate).b).y, 6.0) * (skylight * pi) + shading;
        shading = vec3(0.03) + shading;
        shading = sss * cloud_shadow + shading;

        vec3 emission = col.rgb + torch;

        if (id == 10.0 || id == 11.0 || id == 51.0 || id == 89.0) {
            emission *= sqrt(dot(col.rgb, col.rgb)) * 0.075;
        } else if (id == 50.0) {
            emission *= pow(max(dot(col.rgb, col.rgb) * 0.3 - 0.3, 0.0), 1.0) / 1.0;
        } else if (id == 62.0 || id == 94.0 || id == 149.0) {
                emission *= max(col.r * 5.6 - 0.6, 0.0) * abs(dot(col.rgb, vec3(1.0 / 3.0)) - col.r);
        } else if (id == 76.0 || id == 213.0) {
                emission *= max(col.r * 1.6 - 0.6, 0.0) * 0.1;
        } else if (id == 169.0) {
            emission *= pow(max(dot(col.rgb, col.rgb) * 1.3 - 0.3, 0.0), 2.0) * 0.075;
        } else if (id == 124.0) {
            emission *= sqrt(max(dot(col.rgb, col.rgb) * 1.01 - 0.01, 0.0));
        } else {
            emission *= 0.0;
        }

        return col * shading + emission;
    }
#elif World == 1
    vec3 apply_shading(in vec3 col, in bool do_particles) {
        vec3 shading = vec3(0.0);

        float id = texture(colortex2, texture_coordinate).a * 65535.0;

        float intensity = 2100.0;
        vec3 torch = blackbody(BlocklightTemperature) * intensity;

        vec3 skylight = vec3(0.0);
        skylight = max(ValueFromSphericalHarmonicCoefficientsAlt(shcoeffs, sd.wnormal), 0.0) * mix(pi * 1.5, pi, abs(dot(sd.wnormal, up_vector)));

        if(do_particles && texture(colortex0, texture_coordinate).a < 0.75) skylight = light.sky;

        shading = calculate_blocklight_falloff(decode2x8(texture(colortex2, texture_coordinate).b).x) * torch + shading;
        shading = pow(decode2x8(texture(colortex2, texture_coordinate).b).y, 6.0) * (skylight * pi) + shading;
        shading = vec3(0.5) + shading;

        vec3 emission = col.rgb + torch;

        if (id == 10.0 || id == 11.0 || id == 51.0 || id == 89.0) {
            emission *= sqrt(dot(col.rgb, col.rgb)) * 0.075;
        } else if (id == 50.0) {
            emission *= pow(max(dot(col.rgb, col.rgb) * 0.3 - 0.3, 0.0), 1.0) / 1.0;
        } else if (id == 62.0 || id == 94.0 || id == 149.0) {
                emission *= max(col.r * 5.6 - 0.6, 0.0) * abs(dot(col.rgb, vec3(1.0 / 3.0)) - col.r);
        } else if (id == 76.0 || id == 213.0) {
                emission *= max(col.r * 1.6 - 0.6, 0.0) * 0.1;
        } else if (id == 169.0) {
            emission *= pow(max(dot(col.rgb, col.rgb) * 1.3 - 0.3, 0.0), 2.0) * 0.075;
        } else if (id == 124.0) {
            emission *= sqrt(max(dot(col.rgb, col.rgb) * 1.01 - 0.01, 0.0));
        } else {
            emission *= 0.0;
        }

        return col * shading + emission;
    }
#endif