vec3 reflective_shadow_mapping(vec3 shadow_view_position, vec3 shadow_clip_position, vec3 shadow_view_normal, vec2 noise) {
    const float radius = 4.0;
    const float radius_squared = square(radius);

    const int quality = 6;
    const int samples = square(quality);
    const float per_sample_area = pi * radius_squared / samples;
    const float sample_distance_add = sqrt(per_sample_area / pi);
    const float sgr = quality * 0.5 - 0.5;

    float skylight = decode2x8(texture(colortex2, texture_coordinate).b).y;

    vec3 result;
    vec2 res = textureSize(shadowtex1, 0);
    for(float i = -sgr; i <= sgr; ++i) {
        for(float j = -sgr; j <= sgr; ++j) {
            vec2 offset = (vec2(i, j) + noise - 0.5) / sgr;

            vec3 sample_clip_position = vec3(shadow_clip_position.xy + offset * radius * shadowProjection[0].x, 0.0);
            vec2 sample_coord = distort_shadowspace(sample_clip_position.xy) * 0.5 + 0.5;
            float sample_depth = textureLod(shadowtex0, sample_coord, 0).r;
            sample_clip_position.z = sample_depth * 2.0 - 1.0;
            sample_clip_position.z *= SHADOW_DEPTH_SCALE;

            vec3 sample_view_position = mat3(shadowProjectionInverse) * sample_clip_position + shadowProjectionInverse[3].xyz;

            vec3 sample_vector = shadow_view_position - sample_view_position;
            float distance_squared = dot(sample_vector, sample_vector) + sample_distance_add;
                    sample_vector *= inversesqrt(distance_squared);

            float distance_weight = 1.0 / distance_squared;

            vec3 sample_normal = decodeNormal(textureLod(shadowcolor1, sample_coord, 0).rg * 2.0 - 1.0);

            float directional_weight = saturate(dot(sample_vector, sample_normal)) * saturate(sample_normal.z) / pi;
                  directional_weight *= saturate(dot(-shadow_view_normal,                      sample_vector)) / pi;

            if(directional_weight <= 0.0) { continue; }

            float sample_skylight = textureLod(shadowcolor1, sample_coord, 0).b;
			directional_weight *= saturate(1.0 - 5.0 * abs(sample_skylight - skylight));

            vec3 sample_albedo = srgb_to_linear(textureLod(shadowcolor0, sample_coord, 0).rgb);
            vec3 flux = sample_albedo * directional_weight * distance_weight;

            const float sample_area = 4.0 * radius_squared / samples;

            result += flux * sample_area;
        }
    }
    return pi * result;
}