vec3 hard_shadows_subsurface_scattering(in vec3 position, in bool is_foliage) {
    float bias_mult = calculate_distortion_factor(position.xz);
    float bias = bias_amount * bias_mult / dot(sd.wnormal, shadow_light_vector);
          bias = clamp(bias, 0.0, 0.00015);

    position.z *= rcp(SHADOW_DEPTH_SCALE);

    vec3 ret = vec3(0.0);
    vec2 sample_coord = position.xy;
         sample_coord = distort_shadowspace(sample_coord);
         sample_coord = sample_coord * 0.5 + 0.5;

    float foliage = textureLod(shadowcolor1, sample_coord, 0).a;

    if(foliage > 0.9 && is_foliage) {
        float shadow_depth = textureLod(shadowtex0, sample_coord, 0).r;

        #ifdef ComplexSubsurfaceColor
            vec3 sample_color = srgb_to_linear(textureLod(shadowcolor0, sample_coord, 0).rgb);
        #endif

        vec3 sss_color = vec3(1.0);
        #ifdef ComplexSubsurfaceColor
            sss_color = 1.0 - sample_color;
        #endif

        float sssAmount = SubsurfaceAmount;

        float sss_depth = (SHADOW_DEPTH_RADIUS * max((position.z * 0.5 + 0.5) - shadow_depth, 0)) * sssAmount;

        ret += exp2(-sss_color * sss_depth);
    } else {
        float shadow_opaque = textureLod(shadowtex0, sample_coord, 0).r * 2.0 - 1.0;
        float shadow_translucent = textureLod(shadowtex1, sample_coord, 0).r * 2.0 - 1.0;

        vec3 sample_color = srgb_to_linear(textureLod(shadowcolor0, sample_coord, 0).rgb);
        float sample_alpha = textureLod(shadowcolor0, sample_coord, 0).a;
        vec3 shadow_color_total = mix(vec3(1.0) * mix(vec3(1.0), sample_color, sample_alpha), sample_color, sample_alpha);

        ret += mix(vec3(step(position.z - bias, shadow_opaque)), shadow_color_total, float(step(position.z - bias, shadow_translucent) > step(position.z - bias, shadow_opaque)));
    }

    return ret;
}