vec3 pcss_subsurface_scattering(in vec3 in_position, in bool is_foliage) {
    vec3 position = in_position;
    vec3 ret = vec3(0.0);
    float bias_mult = calculate_distortion_factor(position.xz);
    float bias = bias_amount * bias_mult / dot(sd.wnormal, shadow_light_vector);
          bias = clamp(bias, 0.0, 0.00015);

	const int search_quality = 2;
	const int filter_quality = 2;

	const float search_radius_blocks = 1.0;
	const float light_angular_radius = sun_radius;
    position.z /= SHADOW_DEPTH_SCALE;

	float search_radius = search_radius_blocks * shadowProjection[0].x;

    float noise = fract(frame255 + sin(dot(gl_FragCoord.xy, vec2(18.9898f, 28.633f))) * 4378.5453f) * 5.0;

	float lod = log2(shadowMapResolution * MC_SHADOW_QUALITY * 2 * search_radius * (search_quality * 2 + 1));

	float average_blocker_distance = 0.0;
	float valid_samples = 0.0;
	for (int x = -search_quality; x <= search_quality; ++x) {
        for(int y = -search_quality; y <= search_quality; ++y) {
		    vec2 sample_offset = vec2(x, y) / search_quality;

		    vec2 sample_coord = rotate(noise) * sample_offset * search_radius + position.xy;
		         sample_coord = distort_shadowspace(sample_coord);
		         sample_coord = sample_coord * 0.5 + 0.5;

		    float sample_depth = textureLod(shadowtex0, sample_coord, lod).r * 2.0 - 1.0;

		    if (sample_depth < position.z) {
   		    	average_blocker_distance += position.z - sample_depth;
    	    	valid_samples++;
		    }
        }
	}
	average_blocker_distance /= valid_samples + 1.0;

	average_blocker_distance *= shadowProjectionInverse[2].z;

	float penumbra_size = average_blocker_distance * tan(light_angular_radius);
	penumbra_size *= SHADOW_DEPTH_SCALE;

	float filter_radius = penumbra_size * shadowProjection[0].x;

    lod *= square(filter_radius);
	for (int x = -filter_quality; x <= filter_quality; ++x) {
        for(int y = -filter_quality; y <= filter_quality; ++y) {
		    vec2 sample_offset = vec2(x, y) / filter_quality;

	        vec2 sample_coord = rotate(noise) * sample_offset * filter_radius + position.xy;
	            sample_coord = distort_shadowspace(sample_coord);
	            sample_coord = sample_coord * 0.5 + 0.5;

            float foliage = textureLod(shadowcolor1, sample_coord, 0).a;
    
            if(foliage > 0.9 && is_foliage) {
                float shadow_depth = textureLod(shadowtex0, sample_coord, 0).r;

                #ifdef ComplexSubsurfaceColor
                    vec3 sample_color = srgb_to_linear(textureLod(shadowcolor0, sample_coord, 0).rgb);
                #endif

                vec3 sss_color = vec3(1.0);
                #ifdef ComplexSubsurfaceColor
                    sss_color = 1.0 - sample_color;
                #endif

                float sssAmount = SubsurfaceAmount;

                float sss_depth = (SHADOW_DEPTH_RADIUS * max((position.z * 0.5 + 0.5) - shadow_depth, 0)) * sssAmount;

                ret += exp2(-sss_color * sss_depth);
            } else {
                float shadow_opaque = textureLod(shadowtex0, sample_coord, 0).r * 2.0 - 1.0;
                float shadow_translucent = textureLod(shadowtex1, sample_coord, 0).r * 2.0 - 1.0;

                vec3 sample_color = srgb_to_linear(textureLod(shadowcolor0, sample_coord, 0).rgb);
                float sample_alpha = textureLod(shadowcolor0, sample_coord, 0).a;
                vec3 shadow_color_total = mix(vec3(1.0) * mix(vec3(1.0), sample_color, sample_alpha), sample_color, sample_alpha);

                ret += mix(vec3(step(position.z - bias, shadow_opaque)), shadow_color_total, float(step(position.z - bias, shadow_translucent) > step(position.z - bias, shadow_opaque)));
            }
        }
    }

    return ret / pow(filter_quality * 2 + 1, 2);
}