float minof(vec2 x) { return min(x.x, x.y); }
float minof(vec3 x) { return min(min(x.x, x.y), x.z); }
float minof(vec4 x) { x.xy = min(x.xy, x.zw); return min(x.x, x.y); }

bool raytrace_intersection(
	vec3 start,
	vec3 direction,
	out vec3 position,
	const float qualityConstant,
	const float refinements,
	in float dither
) {
	position   = start;

	start = screenspace_to_viewspace(start, gbufferProjectionInverse);

	direction *= -start.z;
	direction  = viewspace_to_screenspace(direction + start, gbufferProjection) - position;

	vec3 increment = direction * minof((step(0.0, direction) - position) / direction);
	float quality = ceil(qualityConstant * length(increment)) * pi;
	increment /= quality;
	increment *= dither;

	float difference;
	bool  intersected = false;

	for (float i = 0.0; i <= quality && !intersected && position.p < 1.0; i++) {
		position   += increment;
		if (floor(position.st) != vec2(0.0)) break;
		difference  = texture(depthtex1, position.st).r - position.p;
		intersected = difference < 0.0;
	}

	intersected = intersected && (difference + position.p) < 1.0 && position.p > 0.0;

	if (intersected && refinements > 0.0) {
		for (float i = 0.0; i < refinements; i++) {
			increment *= 0.5;
			position  += texture(depthtex1, position.st).r - position.p < 0.0 ? -increment : increment;
			if (any(greaterThan(abs(position - 0.5), vec3(0.5))) || length(increment) * 10 < -difference || texture(depthtex1, position.st).r == 1.0) return false;
		}
	}
	if(!intersected) {
		return false;
	}

	return intersected;
}