vec3 calculate_atmosphere_density(in float h) {
    vec2 rm = exp(-h * inverse_scale_heights + scaled_planet_radius);

    float o = exp(-max(0.0, (35000.0 - h) - atmosphere_radius) * rcp(5000.0))
	            * exp(-max(0.0, (h - 35000.0) - atmosphere_radius) * rcp(15000.0));

    return vec3(rm, o);
}

vec3 calculate_atmosphere_density_analytical(in float b) {
    //This was taken from Jodie's code.
    const vec4 sr = planet_radius + vec4(
        8228.0,
        8228.0 * 1.8 / 8.0,
        35000.0,
        35000.0 + 8228.0
    );
    const float r2 = planet_radius * planet_radius;

    vec4 z = b + sqrt( sr * sr + (b * b - r2) );

    return vec3(z.xy, z.w-z.z);
}

vec3 atmosphere_transmittance(in vec3 ray_vector, in vec3 position, in mat3x3 coeffs) {
    vec3 view_pos = position;
    vec2 ad = raysphere_intersection(view_pos, ray_vector, atmosphere_radius);
    vec2 pd = raysphere_intersection(view_pos, ray_vector, atmosphere_lower_limit);

	bool planet_intersected = pd.y >= 0.0;
   	bool atmosphere_intersected = ad.y >= 0.0;

    vec2 sd = vec2((planet_intersected && pd.x < 0.0) ? pd.y : max(ad.x, 0.0), (planet_intersected && pd.x > 0.0) ? pd.x : ad.y);

	float step_size = length(sd.y - sd.x) * rcp(transmittance_steps);
	vec3  increment = ray_vector * step_size;
	position += ray_vector * sd.x + (increment * 0.5);

    vec3 thickness = vec3(0.0);

    for(int j = 0; j < transmittance_steps; ++j, position += increment) {
        thickness += coeffs * calculate_atmosphere_density(length(position)) * step_size;
    }

    if(planet_intersected) thickness = vec3(10000.0);

    vec3 transmittance = thickness;

    return max(exp(-transmittance), 0.0);
}

#define d0fix(a) ( abs(a) + 1.e-35 )

vec3 absorb_analytical(in vec3 vector, in mat3x3 coeffs) {
    vec3 thickness = calculate_atmosphere_density_analytical(dot(vector, -up_vector) * planet_radius);

    vec3 sky_coeffs = coeffs * thickness;

    vec3 sky_absorb = exp(-sky_coeffs);

    return d0fix(sky_absorb);
}