float clouds_transmittance_density = 0.5;

float cloud_time_3D_large = frameCounter * 0.0075;
float cloud_time_3D_small = frameCounter * 0.0;

float sunlight_visibility_samples = SunlightVisSamples;
float skylight_visibility_samples = SkylightVisSamples;

float get_3D_noise(in vec3 position) {
	float flr = floor(position.z);
	vec2 coord = (position.xy * 0.015625) + (flr * 0.265625);
	vec2 noise = texture(noisetex, coord).xy;
	return mix(noise.x, noise.y, position.z - flr);
}

float fbm3D(in vec3 pos, int octaves, float amplitude, float gain){
    float value = get_3D_noise(pos) * amplitude;
	float inv_weight = 0.0;
    for (int i = 1; i < octaves; i++){
        amplitude *= gain;
		pos *= pi;
		value += get_3D_noise(pos) * amplitude;
		inv_weight += amplitude;
    }
    return value / inv_weight;
}

float fbm3D(in vec3 pos, int octaves){
    float amplitude = 1.0;
    float gain = 0.5;
    return fbm3D(pos, octaves, amplitude, gain);
}

vec3 distort_fbm3D(in vec3 pos, int octaves, float amplitude, float gain){
    vec3 value = vec3(get_3D_noise(pos), get_3D_noise(pos + vec3(3.33, 5.71, 1.96)), get_3D_noise(pos + vec3(7.77, 2.65, 4.37))) * amplitude;
	float inv_weight = 0.0;
    for (int i = 1; i < octaves; i++){
        amplitude *= gain;
		pos *= pi;
		value += vec3(get_3D_noise(pos), get_3D_noise(pos + vec3(3.33, 5.71, 1.96)), get_3D_noise(pos + vec3(7.77, 2.65, 4.37))) * amplitude;
		inv_weight += amplitude;
    }
    return value / inv_weight;
}

vec3 distort_fbm3D(in vec3 pos, int octaves){
    float amplitude = 0.55;
    float gain = 1.0 / 3.0;
    return distort_fbm3D(pos, octaves, amplitude, gain);
}

float clouds_min_altitude = 1600.0;
float clouds_max_altitude = 4000.0;
float cloud_thickness = clouds_max_altitude - clouds_min_altitude;

#define sstep(x, low, high) smoothstep(low, high, x)

float remap(float value, const float originalMin, const float originalMax, const float newMin, const float newMax) {
	return saturate((((value - originalMin) / (originalMax - originalMin)) * (newMax - newMin)) + newMin);
}

float falloff_coverage(float x) {
    return (1.0-sstep(x, clouds_min_altitude, clouds_min_altitude+cloud_thickness*0.1))+sstep(x, clouds_min_altitude+cloud_thickness*0.1, clouds_max_altitude);
}

float get_clouds_density(in vec3 position, in int octaves) {
    if (position.y > clouds_max_altitude || position.y < clouds_min_altitude) return 0.0;

    float height = position.y - clouds_min_altitude;
    float normalize_height = height * (1.0 / cloud_thickness);
    float falloff = saturate(remap(normalize_height, 0.0, 0.4, 0.0, 1.0) * remap(normalize_height, 0.6, 1.0, 1.0, 0.0));
          falloff = sqrt(falloff);
    float coverage_falloff = falloff_coverage(position.y);

    vec3 cloud_pos = position / cloud_thickness;

    vec3 distort = /*0.25 * distort_fbm3D(cloud_pos, 4) +*/ cloud_pos;

    float main_noise = fbm3D(distort + vec3(cloud_time_3D_small, 0.0, cloud_time_3D_small), octaves) * 0.5;
        main_noise  -= 0.55;
        main_noise  -= coverage_falloff*0.5;

    float coverage_factor = 1.0;
    float density  = saturate(main_noise * falloff * coverage_factor * 2.0);

    return density * 0.5;
}

float cloud_od_sunlight(in vec3 pos) {
	vec3 direction = shadow_light_vector;

	float stepsize = cloud_thickness / sunlight_visibility_samples;

	vec3 increment = direction * stepsize;

	float od = 0.0;
	for(float i = 0.0; i < sunlight_visibility_samples; ++i, pos += increment) {
		od += get_clouds_density(pos, NumberOctaves);
	} 

	return od * stepsize * clouds_transmittance_density;
}

float cloud_od_skylight(in vec3 pos) {
	float stepsize = cloud_thickness / skylight_visibility_samples;

	vec3 direction = up_vector;

	vec3 increment = direction * stepsize;

	float od = 0.0;
	for(float i = 0.0; i < skylight_visibility_samples; ++i, pos += increment) {
		od += get_clouds_density(pos, NumberOctaves);
	} 

	return od * stepsize * clouds_transmittance_density;
}

float cloud_od_bouncedlight(in vec3 pos) {
	float stepsize = cloud_thickness / 16;

	vec3 direction = vec3(0, -1, 0);

	vec3 increment = direction * stepsize;

	float od = 0.0;
	for(float i = 0.0; i < 16; ++i, pos += increment) {
		od += get_clouds_density(pos, NumberOctaves);
	}

	return od * stepsize * clouds_transmittance_density;
}

bool distance_is_set = false;
float cloud_distance;

float volumetric_clouds_phase(in float cosTheta) {
	const vec2 g    = vec2(0.25, -0.15);
	const vec2 gm2  = 2.0 * g;
	const vec2 gg   = g * g;
	const vec2 gga1 = 1.0 + gg;
	const vec2 p1   = (0.75 * (1.0 - gg)) / (tau * (2.0 + gg));

	vec2 res = p1 * (cosTheta * cosTheta + 1.0) * pow(gga1 - gm2 * cosTheta, vec2(-1.5));

	return dot(res, vec2(0.4)) + 0.2;
}

const float a = 0.5;
const float b = 0.15;
const float c = 0.85;

vec4 clouds_calculate(in vec3 start_pos, in vec3 end_pos, in bool sky, in float dither, in float steps, in float max_steps) {
	//Assuming I am doing everything right, this should be pretty accurate.

	vec3 world_start = mat3(gbufferModelViewInverse) * start_pos + gbufferModelViewInverse[3].xyz + cameraPosition;
	vec3 direction  = mat3(gbufferModelViewInverse) * normalize(end_pos);

	vec2 distances = vec2(clouds_min_altitude, clouds_max_altitude * 4.0);
	distances = (distances - world_start.y) / direction.y;
	if (distances.y < distances.x) distances = distances.yx;
	distances.x = max(distances.x, 0.0);
	if (distances.y < distances.x) return vec4(0.0, 0.0, 0.0, 1.0);

	float samples = floor(steps * min((distances.y - distances.x) / cloud_thickness, max_steps));

	float view_dot_light = dot(normalize(end_pos), mat3(gbufferModelView) * shadow_light_vector);

	float stepsize = (distances.y - distances.x) / samples;

	vec3 increment = direction * stepsize;
	vec3 pos = world_start + direction * distances.x;

	pos += increment * dither;

	float transmittance = 1.0;
	float optical_depth = 0.0;
	float sky_scatter = 0.0;
	float sun_scatter = 0.0;
	float bun_scatter = 0.0;

	for(int i = 0; i < samples; ++i, pos += increment) {
		float density = get_clouds_density(pos, NumberOctaves);
		if(density <= 0.0) { continue; }
		float step_od = clouds_transmittance_density * stepsize * density;
		float step_transmittance = exp(-step_od);

		if (density > 0.0 && !distance_is_set) {
    		cloud_distance = distance(mat3(gbufferModelViewInverse) * world_start - gbufferModelViewInverse[3].xyz, pos);
    		distance_is_set = true;
		}

		float coeff = transmitted_scattering_integral_vclouds(step_transmittance, clouds_transmittance_density);

		float sun_od = cloud_od_sunlight(pos);
		float sky_od = cloud_od_skylight(pos);
		float bun_od = cloud_od_bouncedlight(pos);

		float powder = mix(1.0 - exp(-sun_od * 2.0), 1.0, view_dot_light * 0.5 + 0.5) * mix(1.0 - exp(-step_od * 1.5), 1.0, view_dot_light * 0.5 + 0.5); //It Just Works.
		float powder_sky = (1.0 - exp(-sky_od * 2.0)) * (1.0 - exp(-step_od * 1.5));
		float powder_bun = (1.0 - exp(-bun_od * 2.0)) * (1.0 - exp(-step_od * 1.5));
		for(int ms = 0; ms < 16; ++ms) {
			float n = float(ms);

			float an = pow(a, n);
			float bn = pow(b, n);
			float cn = pow(c, n);

			float phase = volumetric_clouds_phase(view_dot_light * cn);
			coeff *= an;

			sun_scatter += coeff * powder * phase * exp(-sun_od * bn) * transmittance;
			sky_scatter += coeff * powder_sky * exp(-sky_od * bn) * transmittance;
			bun_scatter += coeff * powder_bun * exp(-bun_od * bn) * transmittance;
		}

		optical_depth += step_od;
		transmittance *= step_transmittance;
	}

	vec4 clouds;
	clouds.rgb = sun_scatter * light.direct * rcp(1.5);
	clouds.rgb += sky_scatter * light.sky * rcp(pi);
	clouds.rgb += bun_scatter * (light.sky * vec3(0.25, 0.3, 0.25) * rcp(pi));
	clouds.rgb *= clouds_transmittance_density;
	clouds.a = transmittance;

	return clouds;
}

float cloud_shadow_calculate(in vec3 start_pos, in vec3 dir, in float steps) {
	vec3 world_start = mat3(gbufferModelViewInverse) * start_pos + gbufferModelViewInverse[3].xyz + cameraPosition;
	vec3 direction  = mat3(gbufferModelViewInverse) * dir;

	float samples = steps;

	vec2 distances = vec2(clouds_min_altitude, clouds_max_altitude);
	distances = (distances - world_start.y) / direction.y;
	if (distances.y < distances.x) distances = distances.yx;
	distances.x = max(distances.x, 0.0);
	if (distances.y < distances.x) return 1.0;

	float stepsize = (distances.y - distances.x) / samples;

	vec3 increment = direction * stepsize;
	vec3 pos = world_start + direction * distances.x;

	pos += increment;

	float transmittance = 1.0;
	float optical_depth = 0.0;

	for(int i = 0; i < samples; ++i, pos += increment) {
		float density = get_clouds_density(pos, NumberOctaves);
		if(density <= 0.0) { continue; }
		float step_od = clouds_transmittance_density * stepsize * density;
		float step_transmittance = exp(-step_od);

		optical_depth += step_od;
		transmittance *= step_transmittance;
	}

	return transmittance;
}