const float clouds_altitude_2d = 700.0;
const float upper_altitude = clouds_altitude_2d + 10.0;
const float lower_altitude = clouds_altitude_2d - 10.0;

const float thickness = upper_altitude - lower_altitude;

float cloud_time_large = frameTimeCounter * 0.0005;
float cloud_time_small = frameTimeCounter / 20.000;

float get_2D_noise(vec2 position) {
	vec2 coord = (position.xy * 0.015625);
	vec2 noise = texture(noisetex, coord).xy;
	return noise.x;
}

float get_clouds_2D_density(vec2 position, float cloudsTime) {
	//Originally from Spectrum, but modified a bit to make the clouds look different.
    position /= clouds_altitude_2d;

	vec2 mainPosition  = (position * 0.15 / 8e-6) * 4e-5 + cloud_time_large;

	const int mainOctaves = 6;

	float mainNoise = get_2D_noise(mainPosition + cloud_time_small);
	for (int i = 1; i < mainOctaves; ++i) {
		mainPosition *= rotateGoldenAngle;
		mainPosition = mainPosition * pi + cloud_time_small;
		mainNoise += get_2D_noise(mainPosition) * exp2(-i);
	} mainNoise = mainNoise * 0.5 + (0.5 * exp2(-mainOctaves));

    float coverage = mix(0.5, 1.0, rainStrength);

	float density  = saturate(mainNoise + coverage - 1.0);
	      density *= density * (3.0 - 2.0 * density);
	      density  = 1.0 - square(1.0 - density);

	return density;
}

const float density = 0.25;

float cloud_od_sunlight_2D(vec3 pos, float od_startpos) {
	vec3 direction = shadow_light_vector;
	float c = 0.8;

	float stepsize = thickness / SunlightVisSamples_2d;
	stepsize /= (direction.y * c + 1 - c);

	vec3 increment = direction * stepsize;
	pos += increment;

	float od = 0.0;
	for(float i = 0.0; i < SunlightVisSamples_2d; ++i, pos += increment) {
		od += get_clouds_2D_density(pos.xz, 0.0);
	} 

	return exp(-density * stepsize * od);
}

float cloud_od_skylight_2D(vec3 pos, float od_startpos) {
	const float stepsize = thickness / SkylightVisSamples_2d;

	vec3 direction = up_vector;

	vec3 increment = direction * stepsize;
	pos += increment;

	float od = 0.0;
	for(float i = 0.0; i < SkylightVisSamples_2d; ++i, pos += increment) {
		od += get_clouds_2D_density(pos.xz, 0.0);
	} 

	return exp(-density * stepsize * od);
}

bool distance_is_set_2d = false;
float cloud_distance_2d;

vec4 clouds(in vec3 view_vector) {
	vec3 eye_position = cameraPosition + gbufferModelViewInverse[3].xyz;
    
	vec2 distances = vec2(lower_altitude, upper_altitude);
	distances = (distances - eye_position.y) / view_vector.y;
	if (distances.y < distances.x) { distances = distances.yx; }
	distances.x = max(distances.x, 0.0);
	if (distances.y < distances.x) { return vec4(vec3(0.0), 1.0); }

	float start_distance = distances.x;
	float end_distance = distances.y;

	vec3 position = eye_position + view_vector * start_distance;

    float od = saturate(get_clouds_2D_density(position.xz, 0.0));

    if (od <= 0.0) { return vec4(vec3(0.0), 1.0); }

    if (od > 0.0 && !distance_is_set_2d) {
        cloud_distance_2d = distance(eye_position - gbufferModelViewInverse[3].xyz, position);
        distance_is_set_2d = true;
    }

    vec3 scattering = vec3(0.0);
    if(od > 0.0) {
        scattering += light.direct * cloud_od_sunlight_2D(position, od) * pi;
        scattering += light.sky * cloud_od_skylight_2D(position, od);
    }

	vec4 clouds = vec4(0.0, 0.0, 0.0, 1.0);
	clouds.rgb = scattering * transmitted_scattering_integral(od, density) * clouds.a;
	clouds.a *= exp(-density * thickness * od);

    return clouds;
}

float cloud_shadow_2d(in vec3 dir, in vec3 eye) {
	vec2 distances = vec2(lower_altitude, upper_altitude);
	distances = (distances - eye.y) / dir.y;
	if (distances.y < distances.x) { distances = distances.yx; }
	distances.x = max(distances.x, 0.0);
	if (distances.y < distances.x) { return 1.0; }

	float start_distance = distances.x;
	float end_distance = distances.y;

	vec3 position = eye + dir * start_distance;

    float od = saturate(get_clouds_2D_density(position.xz, 0.0));

    if (od <= 0.0) { return 1.0; }

    return exp(-density * thickness * od);
}
