struct surface_data {
    vec4 spec;

    vec3 vnormal;
    vec3 wnormal;
    vec3 flat_vnormal;
    vec3 flat_wnormal;

    vec3 eta;
    vec3 etak;

    float f0;
    float roughness;
    float roughness_squared;
    float porosity;
    float sss;
    float emissive;
    float metalness;

    float puddles;
}sd;

vec3 clamp_normal(vec3 normal, vec3 view)  {
    vec3 view_dir = -normalize(view);
    return normalize(normal + (normalize(cross(dFdx(view), dFdy(view))) / pow(saturate(dot(normal, view_dir)) + 0.001, 0.5)));
}

float f0_to_IOR(float f0) {
	f0 = sqrt(f0);
	f0 *= 0.99999; // Prevents divide by 0
	return (1.0 + f0) / (1.0 - f0);
}

float IOR_to_f0(in float ior) {
    float sqrtf0 = (ior - 1.00029) / (ior + 1.00029);
    return square(sqrtf0);
}

surface_data fill_sd_struct(in vec2 coord) {
    surface_data sd;

    float id = texture(colortex2, texture_coordinate).a * 65535.0;

    sd.puddles = texture(colortex5, texture_coordinate).a;

    sd.spec = vec4(decode2x8(texture(colortex5, coord).g), decode2x8(texture(colortex5, coord).b));

    sd.vnormal = mat3(gbufferModelView) * decodeNormal(decode2x8(texture(colortex2, coord).r));
    sd.wnormal = decodeNormal(decode2x8(texture(colortex2, coord).r));
    sd.flat_vnormal = mat3(gbufferModelView) * decodeNormal(decode2x8(texture(colortex5, coord).r));
    sd.flat_wnormal = decodeNormal(decode2x8(texture(colortex5, coord).r));

    sd.f0 = square(sd.spec.g);
    sd.roughness = 1.0 - sd.spec.r;
    sd.roughness_squared = square(sd.roughness);
    sd.porosity = sd.spec.b < (64 / 255) ? sd.spec.b * (255 / 64) : 0.0;
    sd.sss = linear_step(sd.spec.b, 65 / 255, 1.0);
    sd.emissive = sd.spec.a < 254 / 255 ? sd.spec.a : 0;
    sd.metalness = float(sd.spec.g > 0.9);

    if(id == 8.0 || id == 9.0 && isEyeInWater == 0) {
        sd.f0 = 0.02;
        sd.roughness = 1.0 - 1.0;
        sd.roughness_squared = square(sd.roughness);
        sd.porosity = 0.0;
        sd.sss = 0.0;
        sd.emissive = 0;
        sd.metalness = 0.0;
    }

    if(sd.metalness > 0.9) {
        sd.f0 = 1.0;
        //sd.roughness = square(0.01);
    }

    sd.eta = vec3(f0_to_IOR(sd.f0));
    sd.etak = vec3(0.0);

    float iron = 0.0;     //1
    float gold = 0.0;     //2
    float aluminum = 0.0; //3
    float chrome = 0.0;   //4
    float copper = 0.0;   //5
    float lead = 0.0;     //6
    float platinum = 0.0; //7
    float silver = 0.0;   //8

    if(sd.spec.g > 0.900 && sd.spec.g < 0.902) {
        iron = 1.0;     //1
    }
    if(sd.spec.g > 0.902 && sd.spec.g < 0.906) {
        gold = 1.0;     //2
    }
    if(sd.spec.g > 0.906 && sd.spec.g < 0.910) {
        aluminum = 1.0; //3
    }
    if(sd.spec.g > 0.910 && sd.spec.g < 0.913) {
        chrome = 1.0;   //4
    }
    if(sd.spec.g > 0.914 && sd.spec.g < 0.918) {
        copper = 1.0;   //5
    }
    if(sd.spec.g > 0.918 && sd.spec.g < 0.922) {
        lead = 1.0;     //6
    }
    if(sd.spec.g > 0.922 && sd.spec.g < 0.926) {
        platinum = 1.0; //7
    }
    if(sd.spec.g > 0.926 && sd.spec.g < 0.930) {
        silver = 1.0;   //8
    }
 
    //if(coord.x > 0.5) {
        if(iron > 0.9) {
            sd.eta = vec3(2.9114, 2.9497, 2.5845);
            sd.etak = vec3(3.0893, 2.9318, 2.7670);
        }
        if(gold > 0.9) {
            sd.eta = vec3(0.18299, 0.42108, 1.3734);
            sd.etak = vec3(3.4242, 2.3459, 1.7704);
        }
        if(aluminum > 0.9) {
            sd.eta = vec3(1.3456, 0.96521, 0.61722);
            sd.etak = vec3(7.4746, 6.3995, 5.3031);
        }
        if(chrome > 0.9) {
            sd.eta = vec3(3.1071, 3.1812, 2.3230);
            sd.etak = vec3(3.3314, 3.3291, 3.1350);
        }
        if(copper > 0.9) {
            sd.eta = vec3(0.27105, 0.67693, 1.3164);
            sd.etak = vec3(3.6092, 2.6248, 2.2921);
        }
        if(lead > 0.9) {
            sd.eta = vec3(1.9100, 1.8300, 1.4400);
            sd.etak = vec3(3.5100, 3.4000, 3.1800);
        }
        if(platinum > 0.9) {
            sd.eta = vec3(2.3757, 2.0847, 1.8453);
            sd.etak = vec3(4.2655, 3.7153, 3.1365);
        }
        if(silver > 0.9) {
            sd.eta = vec3(0.15943, 0.14512, 0.13547);
            sd.etak = vec3(3.9291, 3.1900, 2.3808);
        }
    //}

    return sd;
}