float smooth_noise(vec2 coord) {
	//return textureSmooth(noisetex, coord / 64.0).r; // slightly faster but has banding artifacts

	vec2 floored = floor(coord);
	vec4 samples = textureGather(noisetex, 0.015625 * floored); // textureGather is slightly offset (at least on nvidia) and this offset can change with driver versions, which is why i floor the coords
	vec4 weights = (coord - floored).xxyy * vec4(1,-1,1,-1) + vec4(0,1,0,1);
	weights *= weights * (-2.0 * weights + 3.0);
	return dot(samples, weights.yxxy * weights.zzww);
}

float almost_identity(float x, float m, float n) {
	if (x > m) return x;
	float t = x / m;
	return (((2.0 * n - m) * t + (2.0 * m - 3.0 * n)) * t * t) + n;
}

float gernsterWaves(vec2 coord, float time, float waveSteepness, float waveAmplitude, float waveLength, vec2 waveDirection){
	const float g = 9.81;
    
	float k = tau / waveLength;
	float w = sqrt(g * k);

	float x = w * time - k * dot(waveDirection, coord);
	float wave = sin(x) * 0.5 + 0.5;

	float h = waveAmplitude * pow(wave, waveSteepness);

	return h;
}

float water_calculate_waves(vec3 pos) {
	/*
	const uint nw = 6u;
	float wave_time = frameTimeCounter * 0.4;
	vec2 p = -(pos.xz + pos.y);

	float waves = 0.0;

	float amplitude = 0.15;

	float steepness = 0.55;

	float w_length = 2.0;

	vec2 dir = vec2(0.25, 0.75);

	float a = tau / 2.618;

	vec2 p_n = (p * 2.5);

	for(uint i = 0u; i < nw; ++i) {
		vec2 noise = vec2(smooth_noise(p_n + i / nw)) * 0.75;

        waves += -gernsterWaves(p + 0.3 * (noise * 2.0 - 1.0) * sqrt(w_length), wave_time, steepness, amplitude, w_length, dir);

		wave_time *= 1.2;

		amplitude *= 0.6;
		steepness *= 1.02;
		w_length *= 0.7;
		dir *= rotate(a);
		p_n *= rotate(a);
	}

    return waves / 6.0;
	*/
	const uint nw = 4u;
	const float wave_height = 0.1;
	float wave_time = frameTimeCounter * 0.15;

	vec2 p = -(pos.xz + pos.y) + wave_time;

	p /= 15.0;

	const float weight_array[nw] = float[nw] (
		1.5,
		8.0,
		10.0,
		25.0
	);

	vec2 p_array[nw] = vec2[nw] (
		wave_time * vec2(0.03, 0.07) + (p / 1.6),
		wave_time * vec2(0.08, 0.06) + (p / 3.0),
		wave_time * vec2(0.07, 0.10) + (p / 4.7),
		wave_time * vec2(0.07, 0.1) + (p / 8.9)
	);

    const vec2 scale_array[nw] = vec2[nw] (
		vec2(4.0, 0.4),
		vec2(1.7, 0.7),
		vec2(1.0, 1.2),
		vec2(2.50, 6.33) / 2.0
	);

	const vec2 inv_scale_array[nw] = vec2[nw] (
		1.0 / scale_array[0],
		1.0 / scale_array[1],
		1.0 / scale_array[2],
		1.0 / scale_array[3]
	);

	vec2 translation_array[nw] = vec2[nw] (
		vec2(p_array[0].y * 0.0, p_array[0].x * 0.2) / scale_array[0],
		vec2(p_array[1].y * 0.2, p_array[1].x * 0.0) / scale_array[1],
		vec2(p_array[2].y * 1.5, p_array[2].x * 1.5) / scale_array[2],
		vec2(p_array[3].y * 1.5, p_array[3].x * 1.7) / scale_array[3]
	);

    bool sharpen[nw] = bool[nw](
        false,
        false,
        false,
        true
    );

	float waves   = 0.0;
	float weights = 0.0;

	for(uint id = 0u; id < nw; id++) {
		float wave = smooth_noise((inv_scale_array[id] * p_array[id] + translation_array[id]) * noiseTextureResolution);
        if(sharpen[id] == true) {
            wave = 1.0 - almost_identity(abs(wave * 2.0 - 1.0), 0.16, 0.08);
        }

		waves   = wave * weight_array[id] + waves;
		weights = weight_array[id] + weights;
	}

	waves /= weights;

	waves *= wave_height;
	waves -= wave_height;

    return waves;
}

vec3 water_normals(vec3 position, in float dist) {
	vec2 diffs = vec2(water_calculate_waves(position + vec3(dist, 0.0, -dist)), water_calculate_waves(position + vec3(-dist, 0.0, dist))) - water_calculate_waves(position - vec3(dist, 0.0, dist));
	vec3 wavenormal = vec3(-2.0 * dist, 4.0 * dist * dist, -2.0 * (dist * dist + dist));
	wavenormal.xz *= diffs;
	
	return wavenormal;
}