float square(in float x) {
    return _square(x);
}

int square(in int x) {
    return _square(x);
}

vec2 square(in vec2 x) {
    return _square(x);
}

vec3 square(in vec3 x) {
    return _square(x);
}

vec4 square(in vec4 x) {
    return _square(x);
}

float pow5(in float x) {
    return _pow5(x);
}

int pow5(in int x) {
    return _pow5(x);
}

vec2 pow5(in vec2 x) {
    return _pow5(x);
}

vec3 pow5(in vec3 x) {
    return _pow5(x);
}

vec4 pow5(in vec4 x) {
    return _pow5(x);
}

float saturate(in float x) {
    return _saturate(x);
}

int saturate(in int x) {
    return _saturate(x);
}

vec2 saturate(in vec2 x) {
    return _saturate(x);
}

vec3 saturate(in vec3 x) {
    return _saturate(x);
}

vec4 saturate(in vec4 x) {
    return _saturate(x);
}

float cubic_smooth(in float x) {
    float y = 3.0, z = 2.0;
    return _cubic_smooth(x, y, z);
}

int cubic_smooth(in int x) {
    int y = 3, z = 2;
    return _cubic_smooth(x, y, z);
}

vec2 cubic_smooth(in vec2 x) {
    vec2 y = vec2(3.0), z = vec2(2.0);
    return _cubic_smooth(x, y, z);
}

vec3 cubic_smooth(in vec3 x) {
    vec3 y = vec3(3.0), z = vec3(2.0);
    return _cubic_smooth(x, y, z);
}

vec4 cubic_smooth(in vec4 x) {
    vec4 y = vec4(3.0), z = vec4(2.0);
    return _cubic_smooth(x, y, z);
}

float rcp(in float x) {
    return _rcp(x, 1.0);
}

float rcp(in int x) {
    return _rcp(x, 1.0);
}

vec2 rcp(in vec2 x) {
    return _rcp(x, 1.0);
}

vec3 rcp(in vec3 x) {
    return _rcp(x, 1.0);
}

vec4 rcp(in vec4 x) {
    return _rcp(x, 1.0);
}

float log10(in float x) {
    return _log10(x, 10.0);
}

int log10(in int x) {
    return int(_log10(x, 10.0));
}

vec2 log10(in vec2 x) {
    return _log10(x, 10.0);
}

vec3 log10(in vec3 x) {
    return _log10(x, 10.0);
}

vec4 log10(in vec4 x) {
    return _log10(x, 10.0);
}

float linear_step(in float x, float low, float high) {
    float data = (x*255.0);
    float mapped = clamp((data-low)/(high-low), 0.0, 255.0);

    return float(mapped/255.0);
}

#define length2(x) dot(x,x)

vec2 raysphere_intersection(vec3 pos, vec3 dir, float radius) {
    float radius_squared = square(radius);
    float pos_times_dir = dot(pos, dir);
    float end_distance = pos_times_dir*pos_times_dir + radius_squared - length2(pos);
    end_distance = sqrt(end_distance);
    
    vec2 ret = -pos_times_dir + vec2(-end_distance, end_distance);
    if(end_distance < 0.0) ret = vec2(-1.0);

    return ret;
}

#include "spectral.glsl"

float bayer2  (vec2 c) { c = 0.5 * floor(c); return fract(1.5 * fract(c.y) + c.x); }
float bayer4  (vec2 c) { return 0.25 * bayer2 (0.5 * c) + bayer2(c); }
float bayer8  (vec2 c) { return 0.25 * bayer4 (0.5 * c) + bayer2(c); }
float bayer16 (vec2 c) { return 0.25 * bayer8 (0.5 * c) + bayer2(c); }
float bayer32 (vec2 c) { return 0.25 * bayer16(0.5 * c) + bayer2(c); }
float bayer64 (vec2 c) { return 0.25 * bayer32(0.5 * c) + bayer2(c); }
float bayer128(vec2 c) { return 0.25 * bayer64(0.5 * c) + bayer2(c); }

float linear_bayer2  (float c) { return fract(c * 0.5); }
float linear_bayer4  (float c) { return 0.5 * linear_bayer2 (c * 0.5) + linear_bayer2(c); }
float linear_bayer8  (float c) { return 0.5 * linear_bayer4 (c * 0.5) + linear_bayer2(c); }
float linear_bayer16 (float c) { return 0.5 * linear_bayer8 (c * 0.5) + linear_bayer2(c); }
float linear_bayer32 (float c) { return 0.5 * linear_bayer16(c * 0.5) + linear_bayer2(c); }
float linear_bayer64 (float c) { return 0.5 * linear_bayer32(c * 0.5) + linear_bayer2(c); }
float linear_bayer128(float c) { return 0.5 * linear_bayer64(c * 0.5) + linear_bayer2(c); }

vec2 sincos(float x) { return vec2(sin(x), cos(x)); }

float hash13(vec3 p3)
{
	p3  = fract(p3 * .1031);
    p3 += dot(p3, p3.yzx + 33.33);
    return fract((p3.x + p3.y) * p3.z);
}

vec2 hash2(vec3 p3) {
	p3 = fract(p3 * vec3(443.897, 441.423, 437.195));
	p3 += dot(p3, p3.yzx + 19.19);
	return fract((p3.xx + p3.yz) * p3.zy);
}

vec3 hash33(vec3 p3)
{
	p3 = fract(p3 * vec3(443.897, 441.423, 437.195));
    p3 += dot(p3, p3.yxz+19.19);
    return fract((p3.xxy + p3.yxx)*p3.zyx);
}

vec4 hash43(vec3 p)
{
	vec4 p4 = fract(vec4(p.xyzx)  * vec4(1031, .1030, .0973, .1099));
    p4 += dot(p4, p4.wzxy+19.19);
    return fract((p4.xxyz+p4.yzzw)*p4.zywx);
}

vec3 gen_unit_vector(vec2 hash) {
    hash.x *= tau; hash.y = hash.y * 2.0 - 1.0;
    return vec3(vec2(sin(hash.x), cos(hash.x)) * sqrt(1.0 - hash.y * hash.y), hash.y);
}

mat2 rotate(float a) {
    vec2 m;
    m.x = sin(a);
    m.y = cos(a);
	return mat2(m.y, -m.x,  m.x, m.y);
}

float shdwcrv (float x){
    x = clamp(x,0.,1.);
    return pow5(x)*(x*(x*(x*(x*70.-315.)+540.)-420.)+126.);
}

float sphere_soft_shadow(const vec3 position, const vec3 L){
    const float k = 10.;
    vec3 oc = position + vec3(0, 1, 0) * 6371e3;
    float b = dot(oc, L);
    float c = dot(oc, oc) - 6371e3*6371e3;
    float h = b*b - c;
    
    float d = -6371e3 + sqrt(max(0.0,6371e3*6371e3-h+max(0.0,h)));
    float t = -b - sqrt(max(0.0,h));
    return (t<0.0) ? 1.0 : shdwcrv(k*d/t);
}

vec2 Rotate(vec2 vector, float angle) {
	vec2 sc = sincos(angle);
	return vec2(sc.y * vector.x + sc.x * vector.y, sc.y * vector.y - sc.x * vector.x);
}
vec3 Rotate(vec3 vector, vec3 axis, float angle) {
	// https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula
	vec2 sc = sincos(angle);
	return sc.y * vector + sc.x * cross(axis, vector) + (1.0 - sc.y) * dot(axis, vector) * axis;
}

vec3 Rotate(vec3 vector, vec3 from, vec3 to) {
	// where "from" and "to" are two unit vectors determining how far to rotate
	// adapted version of https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula

	float cosTheta = dot(from, to);
	if (abs(cosTheta) >= 0.9999) { return cosTheta < 0.0 ? -vector : vector; }
	vec3 axis = normalize(cross(from, to));

	vec2 sc = vec2(sqrt(1.0 - cosTheta * cosTheta), cosTheta);
	return sc.y * vector + sc.x * cross(axis, vector) + (1.0 - sc.y) * dot(axis, vector) * axis;
}

float gaussian(
	in float pos,    // Position
	in float height, // Height
	in float center, // Center
	in float sigma   // Standard deviation
) {
	float p1 = (pos - center) * (pos - center);
	float p2 = 2.0 * (sigma * sigma);
	return height * exp(-(p1 / p2));
}

vec4 textureBicubic(sampler2D sampler, vec2 coord) {
	vec2 res = textureSize(sampler, 0);

	coord = coord * res - 0.5;

	vec2 f = fract(coord);
	coord -= f;

	vec2 ff = f * f;
	vec4 w0;
	vec4 w1;
	w0.xz = 1 - f; w0.xz *= w0.xz * w0.xz;
	w1.yw = ff * f;
	w1.xz = 3 * w1.yw + 4 - 6 * ff;
	w0.yw = 6 - w1.xz - w1.yw - w0.xz;

	vec4 s = w0 + w1;
	vec4 c = coord.xxyy + vec2(-0.5, 1.5).xyxy + w1 / s;
	c /= res.xxyy;

	vec2 m = s.xz / (s.xz + s.yw);
	return mix(
		mix(texture(sampler, c.yw), texture(sampler, c.xw), m.x),
		mix(texture(sampler, c.yz), texture(sampler, c.xz), m.x),
		m.y);
}

vec2 circle_map(in float point) {
    return vec2(cos(point), sin(point));
}