#ifndef MC_GL_RENDERER_RADEON
    layout(location = 0) out vec4 out_color;
    layout(location = 1) out vec4 out_data;
    layout(location = 2) out vec4 out_extra;
    layout(location = 3) out vec4 out_albedo;
#endif

#include "/lib/important.glsl"
#include "/lib/encoders/encoder.glsl"

//Samplers
uniform sampler2D gcolor;
uniform sampler2D specular;
uniform sampler2D normals;
uniform sampler2D noisetex;

//Uniforms
uniform mat4 gbufferModelViewInverse;

uniform float rainStrength, frameTimeCounter;

//Fragment inputs
in mat3 tbn;
in vec3 tint;
in vec3 flat_normal;
in vec3 view_pos;
in vec3 world_pos;
in vec2 texture_coordinate;
in vec2 lightmap_coordinate;
in float ao;
flat in float id;

#ifdef MC_GL_RENDERER_RADEON
    vec4 out_color;
    vec4 out_data;
    vec4 out_extra;
    vec4 out_albedo;
#endif

float smooth_noise(vec2 coord) {
	vec2 floored = floor(coord);
	vec4 samples = textureGather(noisetex, 0.015625 * floored); // textureGather is slightly offset (at least on nvidia) and this offset can change with driver versions, which is why i floor the coords
	vec4 weights = (coord - floored).xxyy * vec4(1,-1,1,-1) + vec4(0,1,0,1);
	weights *= weights * (-2.0 * weights + 3.0);
	return dot(samples, weights.yxxy * weights.zzww);
}

float puddles(in vec3 pos) {
    float puddle =  saturate(pow(smooth_noise(pos.xz / 2) * 2.5 - 0.74, 6.0));
          puddle += saturate(pow(smooth_noise(pos.xz * 2) * 2.5 - 0.74, 6.0)) * puddle;
          puddle += saturate(pow(smooth_noise(pos.xz * 4) * 2.5 - 0.74, 6.0)) * puddle;
	return saturate(puddle);
}

vec3 clamp_normal(vec3 normal, vec3 view)  {
    vec3 view_dir = -normalize(view);
    return normalize(normal + (normalize(cross(dFdx(view), dFdy(view))) / pow(saturate(dot(normal, view_dir)) + 0.001, 0.5)));
}

/* DRAWBUFFERS:0256 */
void main() {
    out_color.rgb = texture(gcolor, texture_coordinate).rgb * tint;
    out_color.rgb *= ao;
    if(texture(gcolor, texture_coordinate).a < 0.1) discard;
    out_color.a = 1.0;

    out_albedo.rgb = texture(gcolor, texture_coordinate).rgb * tint;

    float rain_puddles = puddles(world_pos) * dot(flat_normal, vec3(0, 1, 0));
          rain_puddles = mix(0.0, rain_puddles, rainStrength);

    #ifdef MC_NORMAL_MAP
        vec3 normal = texture(normals, texture_coordinate).rgb * 2.0 - 1.0;
        normal = normalize(tbn * normal);
        normal = clamp_normal(normal, view_pos);
        normal = mat3(gbufferModelViewInverse) * normal;
        normal = mix(normal, flat_normal, rain_puddles);
    #else
        vec3 normal = flat_normal;
    #endif

    out_data.r = encode2x8(encodeNormal(normal));
    out_data.g = 0.0; //This will be used for POM shadows, when I actually add those in.
    out_data.b = encode2x8(bayer16(gl_FragCoord.xy) * (1.0 / 255.0) + (1.0 / 255.0) + lightmap_coordinate);
    out_data.a = id / 65535.0;

    vec4 specular_map = texture(specular, texture_coordinate);

    out_extra.r = encode2x8(encodeNormal(flat_normal));
	#ifdef MC_SPECULAR_MAP
		out_extra.g = encode2x8(specular_map.rg);
		out_extra.b = encode2x8(specular_map.ba);
	#else
		out_extra.g = 0.0;
		out_extra.b = 0.0;
	#endif
	out_extra.a = rain_puddles;

    #ifdef MC_GL_RENDERER_RADEON
        gl_FragData[0] = out_color;
        gl_FragData[1] = out_data;
        gl_FragData[2] = out_extra;
        gl_FragData[3] = out_albedo;
    #endif
}