#ifndef MC_GL_RENDERER_RADEON 
    layout(location = 0) in vec4 vert_pos;
    layout(location = 2) in vec3 vert_normal;
    layout(location = 3) in vec4 vert_color;
    layout(location = 8) in vec2 vert_texcoord;
    layout(location = 9) in vec2 vert_lmcoord;
#else
    vec2 vert_texcoord = gl_MultiTexCoord0.xy;
#endif

#define attribute in
attribute vec4 mc_Entity;
attribute vec4 mc_midTexCoord;
attribute vec4 at_tangent;

#include "/lib/important.glsl"

//Samplers
/* Nothing to see here */

//Uniforms
uniform mat4 gbufferModelViewInverse;
uniform mat4 gbufferModelView;

uniform vec3 cameraPosition;

uniform vec2 taa_offset;

uniform float frameTimeCounter;

//Vertex outputs
out mat3 tbn;
out vec3 tint;
out vec3 flat_normal;
out vec3 view_pos;
out vec3 world_pos;
out vec2 texture_coordinate;
out vec2 lightmap_coordinate;
out float ao;
flat out float id;

float smooth_curve(float x) {
  return x * x * (3.0 - 2.0 * x);
}

float triangle_wave(float x) {
  return abs(fract(x + 0.5) * 2.0 - 1.0);
}

float smooth_triangle_wave(float x) {
  return smooth_curve(triangle_wave(x)) * 2.0 - 1.0;
}

void foliage_wave(inout vec4 position, in bool grass, in bool leaves) {
    if (grass && vert_texcoord.t < mc_midTexCoord.t) {
        float wind = smooth_triangle_wave((frameTimeCounter * pi) * 0.1 + position.x * 0.1 + position.z * 0.001) * 1.0;
        position.x += (smooth_triangle_wave((frameTimeCounter * pi) * 0.3 + position.x * -0.1 + position.z * -0.1) * 2.0 - 1.0) * 0.04 * wind;
        position.y += (smooth_triangle_wave((frameTimeCounter * pi) * 0.1 + position.x * -0.1 + position.y * -0.1) * 2.0 - 1.0) * 0.015 * wind;
        position.y += (smooth_triangle_wave((frameTimeCounter * pi) * 0.1 + position.z * -0.1 + position.y * -0.1) * 2.0 - 1.0) * 0.015 * wind;
        position.z += (smooth_triangle_wave((frameTimeCounter * pi) * 0.1 + position.x * -0.1 + position.z * -0.1) * 2.0 - 1.0) * 0.04 * wind;
    } else if (leaves) {
        float wind = smooth_triangle_wave((frameTimeCounter * pi) * 0.1 + position.x * 0.1 + position.z * 0.001) * 0.5;
        position.x += (smooth_triangle_wave((frameTimeCounter * pi) * 0.1 + position.x * -0.2 + position.z * -0.2) * 2.0 - 1.0) * 0.04 * wind;
        position.y += (smooth_triangle_wave((frameTimeCounter * pi) * 0.1 + position.x * -0.3 + position.y * -0.3) * 2.0 - 1.0) * 0.015 * wind;
        position.y += (smooth_triangle_wave((frameTimeCounter * pi) * 0.1 + position.z * -0.4 + position.y * -0.4) * 2.0 - 1.0) * 0.015 * wind;
        position.z += (smooth_triangle_wave((frameTimeCounter * pi) * 0.1 + position.x * -0.5 + position.z * -0.5) * 2.0 - 1.0) * 0.04 * wind;
    }
}

void main() {
    #ifdef MC_GL_RENDERER_RADEON 
        vec4 pos = gl_Vertex;
    #else
        vec4 pos = vert_pos;
    #endif

    bool grass = false;
    bool leaves = false;
    if(mc_Entity.x == 6.0 || mc_Entity.x == 31.0 || mc_Entity.x == 32.0 || mc_Entity.x == 37.0 || mc_Entity.x == 38.0 || mc_Entity.x == 59.0) grass = true;
    if(mc_Entity.x == 18.0 || mc_Entity.x == 106.0 || mc_Entity.x == 111.0 || mc_Entity.x == 161.0) leaves = true;

    foliage_wave(pos, grass, leaves);

    pos = transMAD(gl_ModelViewMatrix, pos.xyz).xyzz;

    pos = transMAD(gbufferModelViewInverse, pos.xyz).xyzz;

    pos = transMAD(gbufferModelView, pos.xyz).xyzz;
    gl_Position = pos.xyzz * diagonal4(gl_ProjectionMatrix) + vec4(0, 0, gl_ProjectionMatrix[3].z, 0);
    #ifdef TAA
        gl_Position.xy += taa_offset * gl_Position.w;
    #endif

    #ifdef MC_GL_RENDERER_RADEON
    	tbn = mat3(fNormalize(gl_NormalMatrix * at_tangent.xyz), fNormalize(gl_NormalMatrix * cross(at_tangent.xyz, gl_Normal.xyz) * sign(at_tangent.w)), fNormalize(gl_NormalMatrix * gl_Normal.xyz));

        tint = gl_Color.rgb;
        flat_normal = mat3(gbufferModelViewInverse) * fNormalize(gl_NormalMatrix * gl_Normal);
        view_pos = (gl_ModelViewMatrix * gl_Vertex).xyz;
        world_pos = (gbufferModelViewInverse * gl_ModelViewMatrix * gl_Vertex).xyz + cameraPosition;

        texture_coordinate = gl_MultiTexCoord0.st;
        lightmap_coordinate = gl_MultiTexCoord1.st / 255.0;

        ao = gl_Color.a;
    #else
    	tbn = mat3(fNormalize(gl_NormalMatrix * at_tangent.xyz), fNormalize(gl_NormalMatrix * cross(at_tangent.xyz, vert_normal) * sign(at_tangent.w)), fNormalize(gl_NormalMatrix * vert_normal));

        tint = vert_color.rgb;
        flat_normal = mat3(gbufferModelViewInverse) * fNormalize(gl_NormalMatrix * vert_normal);
        view_pos = (gl_ModelViewMatrix * vert_pos).xyz;
        world_pos = (gbufferModelViewInverse * gl_ModelViewMatrix * vert_pos).xyz + cameraPosition;

        texture_coordinate = vert_texcoord;
        lightmap_coordinate = vert_lmcoord / 255.0;

        ao = vert_color.a;
    #endif

    id = mc_Entity.x;
}