#ifndef MC_GL_RENDERER_RADEON
    layout(location = 0) out vec4 out_color;
    layout(location = 1) out vec4 out_data;
    layout(location = 2) out vec4 out_extra;
	layout(location = 3) out vec4 out_albedo;
#endif

#include "/lib/important.glsl"
#include "/lib/encoders/encoder.glsl"

//Samplers
uniform sampler2D gcolor;
uniform sampler2D colortex6;
uniform sampler2D specular;
uniform sampler2D noisetex;

//Uniforms
uniform mat4 gbufferModelViewInverse, gbufferProjectionInverse;
uniform mat4 gbufferProjection;

uniform vec2 taa_offset;

uniform float frameTimeCounter;

//Fragment inputs
in mat3 tbn;
in vec3 tint;
in vec3 flat_normal;
in vec3 view_pos;
in vec3 world_pos;
in vec2 texture_coordinate;
in vec2 lightmap_coordinate;
in float ao;
flat in float id;

#ifdef MC_GL_RENDERER_RADEON
    vec4 out_color;
    vec4 out_data;
    vec4 out_extra;
	vec4 out_albedo;
#endif

#include "/lib/surface/water.glsl"

#include "/lib/converters/converters.glsl"

vec3 water_parallax_mapping(vec3 position, vec3 view_dir) { 
    float min_layers = Water_POM_Quality;
    float num_layers = min_layers;
    float layer_depth = 1.0 / num_layers;
    float cur_layer_depth = 0.0;
    vec3 P = 0.25 * view_dir / abs(view_dir.y); 
    vec3 delta_position = (P / num_layers);

    vec3  cur_pos = vec3(0.0);
    float cur_dm_value = water_calculate_waves(position) * 15.0;

    float prev_dm_value = 0.0;

    for(float i = 0.0; i < min_layers && cur_layer_depth > cur_dm_value; ++i)
    {
        prev_dm_value = cur_dm_value;
        cur_pos = (cur_pos.y - cur_dm_value) * delta_position + cur_pos;
        cur_dm_value = water_calculate_waves(vec3(cur_pos.x, 0.0, cur_pos.z) + position) * 15.0; 
        cur_layer_depth -= layer_depth;
    }

    prev_dm_value /= 15.0;

    vec3 prev_position = cur_pos + delta_position;

    float after_depth  = cur_dm_value - cur_layer_depth;
    float before_depth = prev_dm_value - cur_layer_depth + layer_depth;

    float weight = after_depth / (after_depth - before_depth);
    vec3 final_pos = prev_position * weight + cur_pos * (1.0 - weight);

    //float linear_depth = screenspace_to_viewspace(gl_FragCoord.z, gbufferProjectionInverse) - length(final_pos);
    //gl_FragDepth = viewspace_to_screenspace(linear_depth, gbufferProjection); 

    return vec3(1, 0, 1) * final_pos + position;
	//return position;
}

/* DRAWBUFFERS:1256 */
void main() {
    //gl_FragDepth = gl_FragCoord.z;
    out_color = texture(gcolor, texture_coordinate) * vec4(tint, 1.0);
	out_albedo = out_color;

    vec3 normal = flat_normal;
    if(abs(id - 8.5) < 0.6) {
        normal = mat3(gbufferModelViewInverse) * normalize(tbn * water_normals(water_parallax_mapping(world_pos, normalize(view_pos * tbn).xzy), 0.0005).xzy);
    }

    out_data.r = encode2x8(encodeNormal(normal));
    out_data.g = 0.0;
    out_data.b = encode2x8(lightmap_coordinate);
    out_data.a = id / 65535.0;

    out_extra.r = encode2x8(encodeNormal(flat_normal));
	#ifdef MC_SPECULAR_MAP
		out_extra.g = encode2x8(texture(specular, texture_coordinate).rg);
		out_extra.b = encode2x8(texture(specular, texture_coordinate).ba);
	#else
		out_extra.g = 0.0;
		out_extra.b = 0.0;
	#endif
	out_extra.a = 0.0;

    #ifdef MC_GL_RENDERER_RADEON
        gl_FragData[0] = out_color;
        gl_FragData[1] = out_data;
        gl_FragData[2] = out_extra;
		gl_FragData[3] = out_albedo;
    #endif
}