#ifndef MC_GL_RENDERER_RADEON 
    layout(location = 0) in vec4 vert_pos;
    layout(location = 2) in vec3 vert_normal;
    layout(location = 3) in vec4 vert_color;
    layout(location = 8) in vec2 vert_texcoord;
    layout(location = 9) in vec2 vert_lmcoord;
#endif

#define attribute in
attribute vec4 mc_Entity;
attribute vec4 mc_midTexCoord;
attribute vec4 at_tangent;

#include "/lib/important.glsl"

//Samplers
/* Nothing to see here */

//Uniforms
uniform mat4 gbufferModelViewInverse;
uniform mat4 gbufferModelView;

uniform vec3 cameraPosition;

uniform vec2 taa_offset;

//Vertex outputs
out mat3 tbn;
out vec3 tint;
out vec3 flat_normal;
out vec3 view_pos;
out vec3 world_pos;
out vec2 texture_coordinate;
out vec2 lightmap_coordinate;
out float ao;
flat out float id;

void main() {
    #ifdef MC_GL_RENDERER_RADEON 
        vec4 pos = gl_Vertex;
    #else
        vec4 pos = vert_pos;
    #endif

    pos = transMAD(gl_ModelViewMatrix, pos.xyz).xyzz;

    pos = transMAD(gbufferModelViewInverse, pos.xyz).xyzz;

    pos = transMAD(gbufferModelView, pos.xyz).xyzz;
    gl_Position = pos.xyzz * diagonal4(gl_ProjectionMatrix) + vec4(0, 0, gl_ProjectionMatrix[3].z, 0);
    #ifdef TAA
        gl_Position.xy += taa_offset * gl_Position.w;
    #endif

    #ifdef MC_GL_RENDERER_RADEON
    	tbn = mat3(fNormalize(gl_NormalMatrix * at_tangent.xyz), fNormalize(gl_NormalMatrix * cross(at_tangent.xyz, gl_Normal.xyz) * sign(at_tangent.w)), fNormalize(gl_NormalMatrix * gl_Normal.xyz));

        tint = gl_Color.rgb;
        flat_normal = mat3(gbufferModelViewInverse) * fNormalize(gl_NormalMatrix * gl_Normal);
        view_pos = (gl_ModelViewMatrix * gl_Vertex).xyz;
        world_pos = (gbufferModelViewInverse * gl_ModelViewMatrix * gl_Vertex).xyz + cameraPosition;

        texture_coordinate = gl_MultiTexCoord0.st;
        lightmap_coordinate = gl_MultiTexCoord1.st / 255.0;

        ao = gl_Color.a;
    #else
    	tbn = mat3(fNormalize(gl_NormalMatrix * at_tangent.xyz), fNormalize(gl_NormalMatrix * cross(at_tangent.xyz, vert_normal.xyz) * sign(at_tangent.w)), fNormalize(gl_NormalMatrix * vert_normal.xyz));

        tint = vert_color.rgb;
        flat_normal = mat3(gbufferModelViewInverse) * fNormalize(gl_NormalMatrix * vert_normal);
        view_pos = (gl_ModelViewMatrix * vert_pos).xyz;
        world_pos = (gbufferModelViewInverse * gl_ModelViewMatrix * vert_pos).xyz + cameraPosition;

        texture_coordinate = vert_texcoord;
        lightmap_coordinate = vert_lmcoord / 255.0;

        ao = vert_color.a;
    #endif

    id = mc_Entity.x;
}