layout(location = 0) out vec3 color;
layout(location = 1) out vec3 sky_clear;
layout(location = 2) out vec4 blahblahbufferblah;

#include "/lib/important.glsl"
#include "/lib/color/color_space.glsl"
#include "/lib/encoders/encoder.glsl"

//Samplers
uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex5;
uniform sampler2D colortex6;
uniform sampler2D shadowtex0;
uniform sampler2D shadowtex1;
uniform sampler2D shadowcolor0;
uniform sampler2D shadowcolor1;
uniform sampler2D noisetex;

//Uniforms
uniform mat4 gbufferModelViewInverse, gbufferProjectionInverse;
uniform mat4 gbufferModelView, gbufferProjection;
uniform mat4 shadowModelView, shadowProjection;

uniform vec3 sun_vector, moon_vector, up_vector, shadow_light_vector, sunPosition, upPosition, cameraPosition;

uniform vec2 view_size, view_pixel_size, taa_offset;

uniform ivec2 eyeBrightness;

uniform float frameTimeCounter, rainStrength, eyeAltitude;

uniform int isEyeInWater, frameCounter, frame255;

//Fragment inputs
struct light_col {
    vec3 direct;
    vec3 sky;
};

flat in vec4[3] shcoeffs;
flat in light_col light;
in vec2 texture_coordinate;

const bool colortex0MipmapEnabled = true;

#include "/lib/converters/converters.glsl"
#include "/lib/structs/posdepth_struct.glsl"
#include "/lib/structs/surface_struct.glsl"
#include "/lib/sky/atmos_projection.glsl"
#include "/lib/color/blackbody_functions.glsl"
#include "/lib/color/light_colors.glsl"
#include "/lib/fragment/water_constants.glsl"
#include "/lib/sky/atmosphere/constants.glsl"
#include "/lib/sky/atmosphere/transmittance.glsl"
#include "/lib/sky/atmosphere/phase.glsl"
#include "/lib/sky/clouds_2d.glsl"
#include "/lib/fragment/shading/apply.glsl"
#include "/lib/misc/raytracer.glsl"
#include "/lib/fragment/lens_misc.glsl"
#include "/lib/fragment/gen_bokeh_shape.glsl"

vec3 volume_shadows(in vec3 in_position) {
    vec3 position = in_position;
    vec3 result = vec3(0.0);
    float bias_mult = 1.5 * rcp(calculate_distortion_factor(in_position.xz));
    float bias = bias_amount * bias_mult;

    position.z *= rcp(SHADOW_DEPTH_SCALE);

	vec2 sample_coord = position.xy;
	    sample_coord = distort_shadowspace(sample_coord);
	    sample_coord = sample_coord * 0.5 + 0.5;

    float shadow_translucent = textureLod(shadowtex1, sample_coord, 0).r * 2.0 - 1.0;

    result += vec3(step(position.z - bias, shadow_translucent));

    return result;
}

float fournier_forand_phase(float phi, float mu, float n) {
	// Not sure if this is correct.
	float v = (3.0 - mu) / 2.0;
	float delta = (4.0 / (3.0 * square(n - 1))) * square(sin(phi / 2.0));
	float delta180 = (4.0 / (3.0 * square(n - 1))) * square(sin(pi / 2.0));

	float p1 = 1.0 / (4.0 * pi * square(1.0 - delta) * pow(delta, v));
	float p2 = v * (1.0 - delta) - (1.0 - pow(delta, v)) + (delta * (1.0 - pow(delta, v)) - v * (1.0 - delta)) * pow(sin(phi / 2.0), -2.0);
	float p3 = ((1.0 - pow(delta180, v)) / (16.0 * pi * (delta180 - 1.0) * pow(delta180, v))) * (3.0 * square(cos(phi)) - 1.0);
	return p1 * p2 + p3;
}

vec3 water_fog(in vec3 in_color, in vec3 end, in vec3 start) {
    const int steps_i = WaterVolumeSteps;
	const int stepd_j = 3;
	const int stepd_k = 16;
    const float steps_rcp = rcp(steps_i);

    float phase = fournier_forand_phase(acos(-dot(fNormalize(shadow_light_vector), -fNormalize(viewspace_to_scenespace(end, gbufferModelViewInverse) - gbufferModelViewInverse[3].xyz))), 3.483, 1.08);
    float phase_iso = 0.25 * rcp(pi);
	float phase_sky = 6.0 * phase_iso;
    phase = phase_iso * 0.7 + phase * 0.3;

    vec3 transmittance = vec3(1.0);
    vec3 scattering = vec3(0.0);

    float dither = fract(frameCounter * rcp(31.0) + bayer32(gl_FragCoord.st));

    vec3 world_start = viewspace_to_scenespace(start, gbufferModelViewInverse) + cameraPosition;
    vec3 world_end = viewspace_to_scenespace(end, gbufferModelViewInverse) + cameraPosition;
    vec3 world_increment = (world_end - world_start) * steps_rcp;
    world_start -= world_increment * dither;

    vec3 increment = (end - start) * steps_rcp;
    float step_size = fLength(increment);
    start -= increment * dither;

    mat4 shadow_matrix = shadowProjection * shadowModelView * gbufferModelViewInverse;

    start = mat3(shadow_matrix) * start + shadow_matrix[3].xyz;
    end = mat3(shadow_matrix) * end + shadow_matrix[3].xyz;
    increment = mat3(shadow_matrix) * increment;

    float depth = step_size;

    vec3 sunlight_contribution = light.direct * phase;

    vec4 cur_pos = vec4(start, 1.0);

    vec4 world_pos = vec4(world_start, 1.0);

    vec3 ambient_contribution = vec3(0.0);
    if(isEyeInWater == 1) { 
        ambient_contribution = pow(eyeBrightness.y / 255.0, 6.0) * light.sky; 
    } else {
        ambient_contribution = pow(decode2x8(texture(colortex2, texture_coordinate).b).y, 6.0) * light.sky / 2.0; 
    }
    ambient_contribution *= phase_sky;

    float w_depth = 0.0;
    float fraction = 0.0;
    for(int i = 0; i < steps_i; ++i) {
        cur_pos.xyz += increment;
        world_pos.xyz += world_increment;
        vec3 shadows = mix(volume_shadows(cur_pos.xyz), vec3(0.0), rainStrength);
        vec3 shadow = sunlight_contribution * shadows;
        shadow *= cloud_shadow_2d(shadow_light_vector, world_pos.xyz);
        /*
        if(texture_coordinate.x > 0.5) {
            for(float j = 0.0; j < stepd_j; ++j) {
                vec3 transmittance_multi = vec3(1.0);
                vec3 scattering_multi  = vec3(0.0);
                vec4 cur_pos_multi = cur_pos;
                vec3 increment_multi = mat3(shadow_matrix) * gen_unit_vector(hash2(vec3(gl_FragCoord.xy, j * steps_i + i)));
                vec3 sunlight_contribution_multi = light.direct;
                for(float k = 0.0; k < stepd_k; ++k) {
                    cur_pos_multi.xyz += increment_multi;
                    vec3 shadow_hard_multi = volume_shadows(cur_pos_multi.xyz);

                    vec3 shadow_multi = shadow_hard_multi;
                    shadow_multi *= sunlight_contribution_multi;

                    scattering_multi = shadow_multi * transmittance_multi + scattering_multi;
                    transmittance_multi *= exp(-attenuation_coefficient * length(increment_multi));
                } scattering_multi *= scatter_coefficient;
                scattering_multi *= (1.0 - exp(-attenuation_coefficient * length(increment_multi))) * rcp(attenuation_coefficient);

                scattering += pi * scattering_multi * transmittance_multi * steps_rcp;
            }
        }
        */

        vec3 uod = abs(cur_pos.y - start.y) * attenuation_coefficient;
        vec3 st = exp(-uod * 1.5);

        scattering = (shadow + (ambient_contribution * st)) * transmittance + scattering;
        transmittance *= exp(-attenuation_coefficient * depth);
    } scattering *= scatter_coefficient;
    scattering *= (1.0 - exp(-attenuation_coefficient * depth)) * rcp(attenuation_coefficient);

    return in_color * transmittance + scattering;
}

vec3 water_fog_non_volume(in vec3 in_color, in vec3 end, in vec3 start) {
    vec3 scattering = vec3(0.0);
    vec3 transmittance = vec3(1.0);

    float phase_iso = 0.25 * rcp(pi);
	float phase_sky = 0.01 * phase_iso;

    float depth = distance(end, start);

    vec3 ambient_contribution = vec3(0.0);
    if(isEyeInWater == 1) { 
        ambient_contribution = pow(eyeBrightness.y / 255.0, 6.0) * light.sky; 
    } else {
        ambient_contribution = pow(decode2x8(texture(colortex2, texture_coordinate).b).y, 6.0) * light.sky; 
    }
    ambient_contribution *= phase_sky;

    scattering = (light.direct * ambient_contribution) * transmittance + scattering;
    transmittance *= exp(-attenuation_coefficient * depth);
    scattering *= scatter_coefficient;
    scattering *= (1.0 - exp(-attenuation_coefficient * depth)) * rcp(attenuation_coefficient);

    return in_color * transmittance + scattering;
}

#include "/lib/fragment/brdf.glsl"

vec3 calculate_cloud_fade(mat3 atten_coeff) {
    float cloud_distance_planet = fLength(fNormalize(pd.scene_position)) * fNormalize(pd.scene_position).y * 0.5 + eyeAltitude + planet_radius;
    float rayleigh = exp(-cloud_distance_planet / 8e3 + scaled_planet_radius.x);
    float mie = exp(-cloud_distance_planet / 1.2e3 + scaled_planet_radius.y);
    float ozone = exp(-max(0.0, (35000.0 - cloud_distance_planet) - atmosphere_radius) /  5000.0)
	            * exp(-max(0.0, (cloud_distance_planet - 35000.0) - atmosphere_radius) / 15000.0);
    vec3 rmo = vec3(rayleigh, mie, ozone);

    return atten_coeff * rmo;
}

vec3 raytrace_reflection() {
    //Need to find some optimizations for this.
    vec3 reflectionsky;
    vec3 reflectionssr;

    vec3 ray_direction = normalize(pd.view_position);

    if(sd.f0 <= 0.0) return vec3(0.0);

    int i = 0;
    int rays = 1;

    for(i = 0; i < rays; ++i) {
        vec3 rough_normal = is_ggx(sd.vnormal, hash43(ray_direction + i), sd.roughness_squared);
        if(dot(rough_normal, reflect(ray_direction, rough_normal)) < 0.0) continue; 

        vec3 hit_pos;

        bool hit = raytrace_intersection(pd.screen_position, reflect(ray_direction, rough_normal), hit_pos, 8, 4, 1.0);
        if(hit) {
            reflectionssr = texelFetch(colortex0, ivec2(hit_pos.xy * view_size), 0).rgb * fresnel(sd.eta, sd.etak, dot(reflect(normalize(pd.view_position), rough_normal), rough_normal)) + reflectionssr;
        } else {
            vec4 clouds = clouds(mat3(gbufferModelViewInverse) * reflect(ray_direction, rough_normal));
            vec3 cloud_mix = calculate_cloud_fade(coefficients_extinction);
            vec3 sky = texelFetch(colortex3, ivec2(project_sky(mat3(gbufferModelViewInverse) * reflect(ray_direction, rough_normal)) * view_size), 0).rgb;
                 sky = mix(sky, sky * clouds.a + clouds.rgb, exp(-cloud_mix * cloud_distance_2d));
            reflectionsky = pow(decode2x8(texture(colortex2, texture_coordinate).b).y, 6.0) * 1.45 * sky * fresnel(sd.eta, sd.etak, dot(mat3(gbufferModelViewInverse) * reflect(ray_direction, rough_normal), mat3(gbufferModelViewInverse) * rough_normal)) + reflectionsky;
            continue;
        }
    }
    reflectionsky *= rcp(i);
    reflectionssr *= rcp(i);
    reflectionssr *= MaxColorRange;

    vec3 reflection = reflectionssr + reflectionsky;

    #ifdef SunReflection
        float transparent_alpha = saturate(texture(colortex1, texture_coordinate).a * 100.0);
        bool is_transparent = false;
        if(transparent_alpha == 1.0) is_transparent = true;

        if(is_transparent || sd.metalness > 0.9) {
            float ndotl = dot(mat3(gbufferModelView) * decodeNormal(decode2x8(texture(colortex5, texture_coordinate).r)), mat3(gbufferModelView) * shadow_light_vector);

            if(ndotl < 0.0) {
                vec3 shadows = vec3(0.0);
            } else {
                vec3 shadows = calculate_shadow_map(pd.shadow_clip_position, false);
                shadows *= max(0.0, ndotl);
                shadows *= light.direct;
                shadows *= cloud_shadow_2d(shadow_light_vector, pd.scene_position + cameraPosition);

                reflection += dot(sunPosition, upPosition) < 0.5 ? vec3(0.0) : min(specular_BRDF(-normalize(pd.view_position), sd.vnormal, mat3(gbufferModelView) * sun_vector, sd.roughness_squared), 1e3*1e3) * shadows;
                reflection += dot(sunPosition, upPosition) < 0.5 ? min(specular_BRDF(-normalize(pd.view_position), sd.vnormal, mat3(gbufferModelView) * moon_vector, sd.roughness_squared), 1e3*1e3) * shadows : vec3(0.0);
            }
        }
    #endif

    return max(reflection, 0.0);
}

vec3 blend_metallic(vec3 diffuse, vec3 specular, vec3 albedo, float metalness) {
    return specular * (metalness + (1.0 - metalness)) + (-metalness * diffuse + diffuse);
}

/* DRAWBUFFERS:036 */
void main() {
    pd = fill_pd_struct(texture_coordinate);
    sd = fill_sd_struct(texture_coordinate);

    color = textureLod(colortex0, texture_coordinate, 0).rgb * MaxColorRange;

    float id = texture(colortex2, texture_coordinate).a * 65535.0;

    vec3 transparent_color = srgb_to_linear(texture(colortex1, texture_coordinate).rgb);
    float transparent_alpha = texture(colortex1, texture_coordinate).a;
    if(id == 8.0 || id == 9.0) transparent_alpha = 0.0;

    if(id == 8.0 || id == 9.0 && land_mask(pd.depth0) && isEyeInWater == 0) color = water_fog(color, pd.end_view_position, pd.view_position) * (1.0 - fresnel(sd.eta, sd.etak, dot(reflect(normalize(pd.view_position), sd.vnormal), sd.vnormal)));

    if(land_mask(pd.depth0)) {
        color *= 1.0 - sd.metalness;
        color += raytrace_reflection();
    }

    if(land_mask(pd.depth0) && id != 8.0 || id != 9.0) color = color * mix(vec3(1.0), transparent_color, transparent_alpha);

    if(isEyeInWater == 1) color = water_fog(color, pd.view_position, vec3(0.0));

    sky_clear = vec3(0.0);

    color /= MaxColorRange;

    color = max(color, 0.0);

    vec2 bokehCoord = (texture_coordinate - bokeh_offset) * pow(2.0, LensLOD);
	float bokehAlpha = float(bokehCoord.x < 1.0 && bokehCoord.y < 1.0 && bokehCoord.x > 0.0 && bokehCoord.y > 0.0);

    if(bokehAlpha <= 0.0) {
        blahblahbufferblah = vec4(0.0);
    } else {
        blahblahbufferblah = vec4(bokehImage(LensLOD, bokeh_offset), bokehAlpha);
    }
}