/* DRAWBUFFERS:07 */
layout(location = 0) out vec3 color;
layout(location = 1) out vec4 dof_temporal;

#include "/lib/important.glsl"

//Samplers
uniform sampler2D colortex0;
uniform sampler2D colortex6;
uniform sampler2D colortex7;
uniform sampler2D depthtex0;
uniform sampler2D depthtex1;

//Uniforms
uniform mat4 gbufferProjectionInverse, gbufferProjection;

uniform vec2 view_pixel_size, view_size, taa_offset;

uniform float frameTime, viewWidth, viewHeight, aspectRatio;
uniform float centerDepthSmooth;

//Fragment inputs
in vec2 texture_coordinate;

#include "/lib/converters/converters.glsl"
#include "/lib/fragment/lens_misc.glsl"

const bool colortex0MipmapEnabled = true;

vec2 GetDistOffset(vec2 uv, vec2 pxoffset, float barrel, float anamorphic) {
    vec2 tocenter = uv.xy - 0.5;
    vec3 prep = normalize(vec3(tocenter.y, -tocenter.x, 0.0));

    float angle = fLength(tocenter.xy) * 2.221 * barrel;
    vec3 oldoffset = vec3(pxoffset, 0.0);
    oldoffset.xy *= vec2(anamorphic, 1.0 / anamorphic);

    vec3 rotated = oldoffset * cos(angle) + cross(prep, oldoffset) * sin(angle) + prep * dot(prep, oldoffset) * (1.0 - cos(angle));

    return rotated.xy;
}

vec3 depth_of_field() {
    float r = 1.0;
    mat2 rot = rotate(2.39996322972865332);

    float focus = abs(screenspace_to_viewspace(texture(colortex7, vec2(0.5)).a, gbufferProjectionInverse));
    float depth = abs(screenspace_to_viewspace(texture(depthtex1, texture_coordinate).r, gbufferProjectionInverse));

    float aperture = camera_aperture(0.5 * sensor_width * gbufferProjection[0].x);
    float coc = calculate_camera_coc(focus, depth);
    coc = clamp(coc, 0.0, (1.0/FSTOP)*32.0);
    //coc = pow(fLength(texture_coordinate - 0.5), 2.0) * (aperture * 1e3);
    //coc = (texture_coordinate.y - 0.5) * (aperture * 1e3);

    float samples = 16.0;
    float dof_samples = square(samples);

    vec2 pcoc_angle   = vec2(0.0, coc); // Start angle
    vec2 sample_angle = vec2(0.0, 1.0);
    float size_correct = sqrt(dof_samples) * 1.35914091423;

    float lod = log2(3.0 * view_size.y * coc / sqrt(dof_samples));

    vec3 result = vec3(0.0);
    vec3 weight = vec3(0.0);
    for (float i = 0.0; i < dof_samples; i++) {
        r += 1. / r;
        pcoc_angle = rot * pcoc_angle;
        sample_angle = rot * sample_angle;
        vec2 offset;
        vec3 bokeh = vec3(1.0);

        offset = GetDistOffset(texture_coordinate, (r - 1.0) * sample_angle, 0.0, 1.0) / (size_correct) * 0.5 + 0.5;

        bokeh = texture(colortex6, offset / pow(2.0, LensLOD) + bokeh_offset).rgb;
    
        if(bokeh.r > 0.0 || bokeh.g > 0.0 || bokeh.b > 0.0) {
            offset = GetDistOffset(texture_coordinate, (r - 1.0) * pcoc_angle, DISTORTION_BARREL, DISTORTION_ANAMORPHIC) / (size_correct) * 0.5 * vec2(1.0, aspectRatio) * aperture;

			float coc_blurred = calculate_camera_coc(focus, abs(screenspace_to_viewspace(texture(depthtex1, texture_coordinate + offset).r, gbufferProjectionInverse)));
            coc_blurred = clamp(coc_blurred, 0.0, (1.0/FSTOP)*32.0);

			if ((coc - coc_blurred) > 0.5) continue;

            result += textureLod(colortex0, texture_coordinate + offset, 0).rgb * bokeh;
            weight += bokeh;
        }
    }
    result /= weight;

    return max(result, 0.0);
}

//This calculates the camera shutter speed and the smooth depth.
float camera_shutter_speed() {
    float depth = texture(depthtex1, vec2(0.5)).r;
    float depth_smooth = texture(colortex7, vec2(0.5)).a;
    //Should be doing 1.0 / s, with s being shutterspeed, but that did not cause the speed of the DOF focus to be working right, so I did this instead.
    return mix(depth, depth_smooth, saturate(1.0 - (ShutterSpeed / 255.0)));
}

void main() {
    //color = depth_of_field() * MaxColorRange;
    color = textureLod(colortex0, texture_coordinate, 0).rgb * MaxColorRange;
    color = color / MaxColorRange;
    dof_temporal.rgb = texture(colortex7, texture_coordinate).rgb;
    dof_temporal.a = camera_shutter_speed();
    //dof_temporal.r = 1.0;
}