/* DRAWBUFFERS:04 */
layout(location = 0) out vec3 color;
layout(location = 1) out vec4 temporal;

#include "/lib/important.glsl"

//Samplers
uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex4;

//Uniforms
uniform vec2 view_pixel_size, view_size;

uniform float frameTime;
uniform float viewWidth, viewHeight;

//Fragment inputs
in vec2 texture_coordinate;

const bool colortex0MipmapEnabled = true;

#include "/lib/color/blackbody_functions.glsl"
#include "/lib/color/light_colors.glsl"

float compute_EV100_from_avg_luminance(float avgLuminance) {
    return log2(avgLuminance * 100.0 / 12.5);
}

float convert_EV100_to_exposure (float EV100) {
    float maxLuminance = 1.2 * pow(2.0, EV100);

    return maxLuminance;
}

vec2 get_diaphragm_shape(in vec2 o, in float n, in float f, in float shutter_max) {
   vec2 ab = (o * 2.0) - vec2(1.0);
   vec2 shape = ((ab.x * ab.x) > (ab.y * ab.y)) ? vec2((abs(ab.x) > 1e-8) ? (pi * ab.y) / (4.0 * ab.x) : 0.0, ab.x) : vec2((abs(ab.y) > 1e-8) ? (pi * 0.5) - ((pi * ab.x) / (4.0 * ab.y)) : 0.0, ab.y);
   shape.x += f * shutter_max;
   shape.y *= (f > 0.0) ? pow((cos(pi / n) / cos(shape.x - ((tau / n) * floor(((n * shape.x) + pi) / tau)))), f) : 1.0;
   return vec2(cos(shape.x), sin(shape.x)) * shape.y;
}

float camera_exposure() {
    vec3 average_samples = textureLod(colortex0, vec2(0.5), log2(max(view_size.x, view_size.y))).rgb;

    float average_brightness = 1.5 * rcp(dot(average_samples, vec3(2.0 / 0.04)));

    float auto_ev_100 = compute_EV100_from_avg_luminance(clamp(average_brightness, 0.04, 40.0));
    float current_ev = auto_ev_100;
    float exposure = convert_EV100_to_exposure(current_ev);
    float exposure_time = frameTime * rcp(mix(3.5, 0.35, float(exposure < texture(colortex4, vec2(0.0) + view_pixel_size * 0.5).a)) + frameTime);
    exposure = mix(texture(colortex4, vec2(0.0) + view_pixel_size * 0.5).a, exposure, exposure_time);

    return max(exposure, 0.0);
}

void main() {
    color = textureLod(colortex0, texture_coordinate, 0).rgb * MaxColorRange;
    float exposure = camera_exposure();
    color /= MaxColorRange;

    temporal = vec4(texture(colortex4, texture_coordinate).rgb, exposure);
}