/* DRAWBUFFERS:0 */
layout(location = 0) out vec3 out_color;

#include "/lib/important.glsl"

//Samplers
uniform sampler2D colortex0;
uniform sampler2D colortex3;
uniform sampler2D colortex4;
uniform sampler2D colortex6;
uniform sampler2D depthtex0;
uniform sampler2D depthtex1;
uniform sampler2D noisetex;

//Uniforms
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 gbufferModelViewInverse;
uniform mat4 gbufferPreviousModelView, gbufferPreviousProjection;

uniform vec3 sunPosition, cameraPosition, previousCameraPosition;

uniform vec2 view_pixel_size, view_size, taa_offset;

uniform float frameTimeCounter, aspectRatio, centerDepthSmooth;

uniform int isEyeInWater;

//Fragment inputs
in vec2 texture_coordinate;

#include "/lib/converters/converters.glsl"

//#define MotionBlur

vec3 motion_blur() {
	vec4 result = texture(colortex0, texture_coordinate) * MaxColorRange;
	#ifdef MotionBlur
		result = vec4(0.0);
		vec4 position = vec4(texture_coordinate, texture(depthtex0, texture_coordinate).r, 1.0) * 2.0 - 1.0;

		vec4 previous = gbufferProjectionInverse * position; previous /= previous.w;
		previous = gbufferModelViewInverse * previous;
		previous.xyz += cameraPosition - previousCameraPosition;
		previous = gbufferPreviousModelView * previous;
		previous = gbufferPreviousProjection * previous; previous /= previous.w;

		vec2 velocity = (previous - position).xy;
		velocity = velocity/(1.0+length(velocity))*1.5*0.02;

		vec2 blurVec = velocity * 10.0;

		for (int i = 1; i < 64; ++i) {
			vec2 offset = blurVec * (float(i) / float(64 - 1) - 0.5);

			result += texture(colortex0, texture_coordinate + offset);
		}

		result /= float(64);
		result = result * MaxColorRange;
	#endif
	return result.rgb;
}

void main() {
    out_color = motion_blur() / MaxColorRange;
}