layout(location = 0) out vec3 out_color;

/*
const int colortex0Format = RGBA16F; //Scene color
const int colortex1Format = RGBA16F; //Transparent color
const int colortex2Format = RGBA16; //Gbuffer data
const int colortex3Format = R11F_G11F_B10F; //Sky color
const int colortex4Format = RGBA16F; //Temporal data
const int colortex5Format = RGBA16; //"Extra" gbuffer data
const int colortex6Format = RGBA16F;
const int colortex7Format = RGBA32F;
*/

#include "/lib/important.glsl"
#include "/lib/color/color_space.glsl"

//Samplers
uniform sampler2D colortex0;
uniform sampler2D colortex3;
uniform sampler2D colortex4;
uniform sampler2D colortex6;
uniform sampler2D colortex7;
uniform sampler2D depthtex0;
uniform sampler2D depthtex1;
uniform sampler2D noisetex;

//Uniforms
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 gbufferModelViewInverse;
uniform mat4 gbufferPreviousModelView, gbufferPreviousProjection;

uniform vec3 sunPosition, cameraPosition, previousCameraPosition;

uniform vec2 view_pixel_size, view_size, taa_offset;

uniform float frameTimeCounter, aspectRatio, centerDepthSmooth;

uniform int isEyeInWater;

//Fragment inputs
in vec2 texture_coordinate;

#include "/lib/converters/converters.glsl"
#include "/lib/color/tonemapping.glsl"
#include "/lib/fragment/lens_misc.glsl"

void apply_bloom(inout vec3 in_color) {
    vec2 px = view_pixel_size;
	vec3 bloom  = textureBicubic(colortex3, (texture_coordinate / exp2(1)) + vec2(0.00000           , 0.00000            )).rgb * 0.475;
	bloom += textureBicubic(colortex3, (texture_coordinate / exp2(2)) + vec2(0.00000           , 0.50000 + px.y * 19)).rgb * 0.625;
	bloom += textureBicubic(colortex3, (texture_coordinate / exp2(3)) + vec2(0.25000 + px.x * 2, 0.50000 + px.y * 19)).rgb * 0.750;
	bloom += textureBicubic(colortex3, (texture_coordinate / exp2(4)) + vec2(0.25000 + px.x * 2, 0.62500 + px.y * 37)).rgb * 0.850;
	bloom += textureBicubic(colortex3, (texture_coordinate / exp2(5)) + vec2(0.31250 + px.x * 4, 0.62500 + px.y * 37)).rgb * 0.925;
	bloom += textureBicubic(colortex3, (texture_coordinate / exp2(6)) + vec2(0.31250 + px.x * 4, 0.65625 + px.y * 55)).rgb * 0.975;
	bloom += textureBicubic(colortex3, (texture_coordinate / exp2(7)) + vec2(0.46875 + px.x * 6, 0.65625 + px.y * 55)).rgb * 1.000;
	bloom /= pi;
	bloom /= MaxColorRange;

    float a = 0.25 * 0.25;
	if(isEyeInWater == 1) {
		a = 0.4;
	}
    in_color = mix(in_color, bloom, a);
}

vec3 saturation(vec3 working_color)
{
	float grey = dot(lumacoeff_rec709, working_color);
	vec3 ret = grey + (1.0 + SaturationMod) * (working_color - grey);
	
	return ret;
}

float contrast(float x, float eps, float logMidpoint, float contrast)
{
	float logX = log2(x+eps);
	float adjX = logMidpoint + (logX - logMidpoint) * (1.0 + contrast);
	float ret = max(0.0f, exp2(adjX) - eps);
	return ret;
}

vec3 contrast(in vec3 in_color) {
	return vec3(
        contrast(in_color.x, 1e-5, log2(0.18), ContrastMod), 
        contrast(in_color.y, 1e-5, log2(0.18), ContrastMod), 
        contrast(in_color.z, 1e-5, log2(0.18), ContrastMod)
    );
}

vec3 vibrance(in vec3 in_color) {
    float lum = dot(in_color, lumacoeff_rec709);
    vec3 mask = (in_color - vec3(lum));
    mask = clamp(mask, 0.0, 1.0);
    float lum_mask = dot(lumacoeff_rec709, mask);
    lum_mask = 1.0 - lum_mask;
    return mix(vec3(lum), in_color, (1.0 + VibrancyMod) * lum_mask);
}    

vec3 color_correct(in vec3 working_color) {
	working_color = white_balance(working_color);
    working_color = vibrance(working_color);
    working_color = saturation(working_color);
	working_color = contrast(working_color);
    return working_color;
}

vec3 film_grain(in vec3 color) {
    float noise = fract(frameTimeCounter + sin(dot(gl_FragCoord.xy, vec2(18.9898f, 28.633f))) * 4378.5453f) * 5.0;
	noise /= MaxColorRange;
	return color + (noise / 120.0);
}

vec3 lowlight_desaturate(in vec3 color) {
    const vec3 rod_response = vec3(0.15, 0.50, 0.35);

    float rod_color = dot(color * rod_response, vec3(1.0));
    float desaturated = dot(color, rod_response);
    color = mix(color, vec3(rod_color), exp2(-4600.0 * desaturated));

    return color;
}

#include "/lib/fragment/gen_bokeh_shape.glsl"
#include "/lib/fragment/lens_system.glsl"
#include "/lib/fragment/lens_flare.glsl"

void main() {
    fill_lens_system();

    out_color = texture(colortex0, texture_coordinate).rgb * 2.0;
    #if World == 1
        out_color *= 2.0;
    #endif

    apply_bloom(out_color);

    #if World == 0
	    out_color += lens_flares();
    #endif

	out_color = film_grain(out_color);

	out_color = lowlight_desaturate(out_color);

	out_color *= texture(colortex4, texture_coordinate).a;

    out_color = tonemap(out_color);
    out_color = color_correct(out_color);
    #if Tonemap != 1 && Tonemap != 0
        out_color = linear_to_srgb(out_color);
    #endif

    out_color += bayer128(gl_FragCoord.st) / 128.0;
}