/* DRAWBUFFERS:04 */
layout(location = 0) out vec3 color;
layout(location = 1) out vec4 temporal;

//Samplers
uniform sampler2D colortex0;
uniform sampler2D colortex4;
uniform sampler2D depthtex1;

//Uniforms
uniform mat4 gbufferPreviousModelView, gbufferPreviousProjection;
uniform mat4 gbufferProjectionInverse, gbufferModelViewInverse;

uniform vec3 cameraPosition;
uniform vec3 previousCameraPosition;

uniform vec2 view_pixel_size, view_size;

uniform float frameTime;

//Fragment inputs
in vec2 texture_coordinate;

#include "/lib/important.glsl"

vec4 reproject(vec4 position) {
	position = position * 2.0 - 1.0;

	position      = gbufferModelViewInverse * gbufferProjectionInverse * position;
	position     /= position.w;
	position.xyz += cameraPosition;

	position.xyz -= previousCameraPosition;
	position      = gbufferPreviousProjection * gbufferPreviousModelView * position;
	position     /= position.w;

	position = position * 0.5 + 0.5;
	return position;
}

vec4 temporal_anti_aliasing() {
    vec4 return_color = vec4(0.0);
	vec4 position = vec4(texture_coordinate, texture(depthtex1, texture_coordinate).r, 1.0);
	vec4 reprojected_pos = reproject(position);

	float blend_weight = 0.9;

	const float offcenter_rejection = 0.5;
	vec2 pixel_center_dist = 1.0 - abs(2.0 * fract(reprojected_pos.xy * view_size) - 1.0);
	blend_weight *= sqrt(pixel_center_dist.x * pixel_center_dist.y) * offcenter_rejection + (1.0 - offcenter_rejection);

	if (clamp(reprojected_pos.xy, 0.0, 1.0) != reprojected_pos.xy) { blend_weight *= 0.0; }

	vec4 curr_frame = texture(colortex0, position.xy);
	vec4 prev_frame = vec4(texture(colortex4, reprojected_pos.xy).rgb, 1.0);

	vec3 tl = texture(colortex0, view_pixel_size * vec2(-1,-1) + texture_coordinate).rgb;
	vec3 tc = texture(colortex0, view_pixel_size * vec2( 0,-1) + texture_coordinate).rgb;
	vec3 tr = texture(colortex0, view_pixel_size * vec2( 1,-1) + texture_coordinate).rgb;
	vec3 ml = texture(colortex0, view_pixel_size * vec2(-1, 0) + texture_coordinate).rgb;
	vec3 mc = curr_frame.rgb;
	vec3 mr = texture(colortex0, view_pixel_size * vec2( 1, 0) + texture_coordinate).rgb;
	vec3 bl = texture(colortex0, view_pixel_size * vec2(-1, 1) + texture_coordinate).rgb;
	vec3 bm = texture(colortex0, view_pixel_size * vec2( 0, 1) + texture_coordinate).rgb;
	vec3 br = texture(colortex0, view_pixel_size * vec2( 1, 1) + texture_coordinate).rgb;

	vec3 min_col = min(min(min(min(min(min(min(min(tl, tc), tr), ml), mc), mr), bl), bm), br);
	vec3 max_col = max(max(max(max(max(max(max(max(tl, tc), tr), ml), mc), mr), bl), bm), br);

	return_color.rgb = mix(curr_frame.rgb, clamp(prev_frame.rgb, min_col, max_col), blend_weight);

    return return_color;
}

void main() {
    vec4 taa = temporal_anti_aliasing();
	#ifdef TAA
		color = (taa.rgb);
	#else
		color = texture(colortex0, texture_coordinate).rgb;
	#endif
	temporal = vec4(taa.rgb, texture(colortex4, texture_coordinate).a);
}