#version 420 compatibility

#ifndef MC_GL_RENDERER_RADEON 
    layout(location = 0) in vec4 vert_pos;
    layout(location = 2) in vec3 vert_normal;
    layout(location = 3) in vec4 vert_color;
    layout(location = 8) in vec2 vert_texcoord;
    layout(location = 9) in vec2 vert_lmcoord;
#else
    vec4 vert_pos = gl_Vertex;
    vec4 vert_color = gl_Color;
    vec3 vert_normal = gl_Normal;
    vec2 vert_texcoord = gl_MultiTexCoord0.xy;
    vec2 vert_lmcoord = gl_MultiTexCoord1.xy;
#endif

#define attribute in
attribute vec4 mc_Entity;
attribute vec4 mc_midTexCoord;
attribute vec4 at_tangent;

#include "/lib/important.glsl"

//Samplers
/* Nothing to see here */

//Uniforms
uniform mat4 shadowModelViewInverse;
uniform mat4 shadowModelView, shadowProjection;

uniform float frameTimeCounter;

//Vertex outputs
out vec3 tint;
out vec3 shadow_normal;
out vec2 texture_coordinate;
out float skylight;
flat out float id;

#include "/lib/distortion.glsl"

float smooth_curve(float x) {
  return x * x * (3.0 - 2.0 * x);
}

float triangle_wave(float x) {
  return abs(fract(x + 0.5) * 2.0 - 1.0);
}

float smooth_triangle_wave(float x) {
  return smooth_curve(triangle_wave(x)) * 2.0 - 1.0;
}

void foliage_wave(inout vec4 position, in bool grass, in bool leaves) {
    if (grass && vert_texcoord.t < mc_midTexCoord.t) {
        float wind = smooth_triangle_wave((frameTimeCounter * pi) * 0.1 + position.x * 0.1 + position.z * 0.001) * 1.0;
        position.x += (smooth_triangle_wave((frameTimeCounter * pi) * 0.3 + position.x * -0.1 + position.z * -0.1) * 2.0 - 1.0) * 0.04 * wind;
        position.y += (smooth_triangle_wave((frameTimeCounter * pi) * 0.1 + position.x * -0.1 + position.y * -0.1) * 2.0 - 1.0) * 0.015 * wind;
        position.y += (smooth_triangle_wave((frameTimeCounter * pi) * 0.1 + position.z * -0.1 + position.y * -0.1) * 2.0 - 1.0) * 0.015 * wind;
        position.z += (smooth_triangle_wave((frameTimeCounter * pi) * 0.1 + position.x * -0.1 + position.z * -0.1) * 2.0 - 1.0) * 0.04 * wind;
    } else if (leaves) {
        float wind = smooth_triangle_wave((frameTimeCounter * pi) * 0.1 + position.x * 0.1 + position.z * 0.001) * 0.5;
        position.x += (smooth_triangle_wave((frameTimeCounter * pi) * 0.1 + position.x * -0.2 + position.z * -0.2) * 2.0 - 1.0) * 0.04 * wind;
        position.y += (smooth_triangle_wave((frameTimeCounter * pi) * 0.1 + position.x * -0.3 + position.y * -0.3) * 2.0 - 1.0) * 0.015 * wind;
        position.y += (smooth_triangle_wave((frameTimeCounter * pi) * 0.1 + position.z * -0.4 + position.y * -0.4) * 2.0 - 1.0) * 0.015 * wind;
        position.z += (smooth_triangle_wave((frameTimeCounter * pi) * 0.1 + position.x * -0.5 + position.z * -0.5) * 2.0 - 1.0) * 0.04 * wind;
    }
}

void main() {
    vec4 pos = vert_pos;

    bool grass = false;
    bool leaves = false;
    if(mc_Entity.x == 6.0 || mc_Entity.x == 31.0 || mc_Entity.x == 32.0 || mc_Entity.x == 37.0 || mc_Entity.x == 38.0 || mc_Entity.x == 59.0) grass = true;
    if(mc_Entity.x == 18.0 || mc_Entity.x == 106.0 || mc_Entity.x == 111.0 || mc_Entity.x == 161.0) leaves = true;

    foliage_wave(pos, grass, leaves);

    pos.xyz = mat3(gl_ModelViewMatrix) * pos.xyz + gl_ModelViewMatrix[3].xyz;

    gl_Position = vec4(shadowProjection[0].x, shadowProjection[1].y, shadowProjection[2].zw) * pos.xyzz + shadowProjection[3];
    gl_Position.xyz = distort_shadowspace(gl_Position.xyz);

    tint = vert_color.rgb;
    shadow_normal = gl_NormalMatrix * vert_normal;

    texture_coordinate = vert_texcoord;
    skylight = vert_lmcoord.y / 255.0;

    id = mc_Entity.x;
}